import os
import subprocess
import sys

_ROOT = os.path.abspath(os.path.dirname(__file__))
key_sleep_script = _ROOT + "/tools/g5k_set_pub_key_sleep.sh"


def g5k_get_seed_store(
    ctx, url="http://public.grenoble.grid5000.fr/~orichard/seed-nix-store.tgz"
):
    store_path = f"{os.environ['HOME']}/.local/share/nix/root/nix/store"
    if len(os.listdir(store_path)) == 0:
        subprocess.call(
            f"wget -c {url} -O - | tar -C  ~/.local/share/nix/root -xz", shell=True
        )
    else:
        ctx.elog(f"Store path and direct parent must be empty: {store_path} ")
        sys.exit(1)



import yaml


class DefaultRole:
    def __init__(self, nb_min_nodes=0):
        self.nb_min_nodes = int(nb_min_nodes)


def default_role_constructor(loader, node):
    nb_min_nodes = loader.construct_scalar(node)
    return DefaultRole(nb_min_nodes)


def get_nxc_loader():
    loader = yaml.SafeLoader
    loader.add_constructor("!DefaultRole", default_role_constructor)
    return loader



import click
import re
from ..context import pass_context  # , on_started, on_finished
from ..actions import read_deployment_info, connect_tmux
from ..flavours import get_flavour_by_name


@click.command("connect")
@click.option("-l", "--user", default="root")
@click.option(
    "-g",
    "--geometry",
    help='Tmux geometry, 2 splitting indications are supported: +/*, examples: "1+3+2" (3 adjacent panes respectively horizontally splited by 1,3 and 2), "2*3" (2 adjacent panes horizontally splitted by 3)',
)
@click.option(
    "-d",
    "--deployment-file",
    help="Deployment file, take the latest created in deploy directory by default",
)
@click.option(
    "-f",
    "--flavour",
    help="flavour, by default it's extracted from deployment file name",
)
@click.option(
    "-i",
    "--identity-file",
    type=click.STRING,
    help="path to the ssh public private used to connect to the deployments",
)
@click.option(
    "-a",
    "--tag",
    help="tagname of the deployment we want to connect on",
)
@click.option("-pc", "--pane-console", is_flag=True, help="Add a pane console")
@click.argument("host", nargs=-1)
@pass_context
# TODO @on_finished(lambda ctx: ctx.state.dump())
# TODO @on_started(lambda ctx: ctx.assert_valid_env())
def cli(
    ctx, user, host, geometry, pane_console, deployment_file, flavour, identity_file, tag
):
    """
    Opens one or more terminal sessions into the deployed nodes. By default, it will connect to all nodes, but we can specify which ones to connect to.

    To connect to several machines at once `nxc` use a tmux (terminal multiplexer) session. Feel free to refer to the [tmux documentation](https://github.com/tmux/tmux/wiki) (or its [cheatsheet](https://tmuxcheatsheet.com/)), especially for the shortcuts to navigate between the different tabs.

    ## Examples

    - `nxc connect`

        Open a Tmux session with one panel for each node.

    - `nxc connect server`

        Connect to the `server` node. It runs on the current shell (Tmux is not used in this case)
    """
    read_deployment_info(ctx, deployment_file, flavour, tag)

    # determine flavour name
    if not flavour:
        match = re.match(r"^.*?::([^:]+?)(?:::|\.).*$", ctx.deployment_filename)
        if match:
            flavour = match.group(1)
        else:
            raise click.ClickException(
                "Cannot determined flavour, one must by provided (option fix deployment file name or use --flavour option"
            )

    ctx.flavour = get_flavour_by_name(flavour)(ctx)

    if not host or len(host) > 1:
        # TODO  add wait_ssh
        connect_tmux(ctx, user, host, identity_file, pane_console, geometry, "nxc")
    else:
        ctx.flavour.ext_connect(user, host[0], ssh_key_file=identity_file)

        # if "docker-compose-file" in ctx.deployment_info:
        #    connect_docker(ctx, user, host[0])
        # else:
        #    connect(ctx, user, host[0])



#!/usr/bin/env python
import argparse
import json
import os
import os.path as op
import sys


def get_ssh_pub_key(ssh_pub_key_file):
    if not ssh_pub_key_file:
        ssh_pub_key_file = os.environ["HOME"] + "/.ssh/id_rsa.pub"
    with open(ssh_pub_key_file, "r") as f:
        sshkey_pub = f.read().rstrip()
    return sshkey_pub


def realpath_from_store(path, include_prefix_store=False):
    p = op.realpath(path)
    for store_path in [
        f"{os.environ['HOME']}/.local/share/nix/root/nix",
        f"{os.environ['HOME']}/.nix",
    ]:
        new_p = f"{store_path}{p[4:]}"
        if op.exists(new_p):
            if include_prefix_store:
                return new_p, store_path
            else:
                return new_p
    if op.exists(p):
        if include_prefix_store:
            return p, None
        else:
            return p
    print(f"{path} does not exist in standard store or alternate", file=sys.stderr)
    sys.exit(1)


def read_role_distribution(role_distribution_file):
    with open(role_distribution_file, "r") as f:
        role_distribution = json.load(f)
    return role_distribution


def nested_deployment(composition, role_distribution, hostbase_filter=None):
    # import pdb; pdb.set_trace()
    deployment_info = {}
    ip_hosts = []
    route_commands = []
    net_index = 0
    host_index = 1
    network = ""

    for hostbase, v in role_distribution["role_distribution"].items():
        nest_host_index = 1
        l, h = v["range"].split("-")
        for host_index in range(int(l), int(h) + 1):
            host = f"{hostbase}{host_index}"
            role = v["role"]
            folding = v["folding"]
            toplevel = composition[role]["toplevel"]

            print(f"host: {host}, role: {role}, folding: {folding}")
            for i in range(1, int(folding) + 1):
                ip = f"10.0.{3 + net_index}.{1 + i}"
                nested_host = f"{role}{nest_host_index}"
                print(f"    {nested_host} : {ip}")
                nest_host_index += 1
                if not hostbase_filter or (hostbase_filter == host):
                    if hostbase_filter:
                        print(f"{host}:")
                    deployment_info[ip] = {
                        "role": role,
                        "host": nested_host,
                        "toplevel": toplevel,
                    }
                ip_hosts.append(f"{ip} {nested_host}")

            if hostbase_filter:
                if hostbase_filter != host:
                    route_commands.append(
                        f"ip route add 10.0.{3 + net_index}.0/24 via $(getent hosts {host} | cut -d' ' -f1)"
                    )
                else:
                    network = f"10.0.{3 + net_index}.0/24"
            net_index += 1
    return deployment_info, ip_hosts, route_commands, network


# role_distribution = read_role_distribution("nested_role_distribution.json")

# # deployment = nested_deployment("toplevel")
# # json_formatted_str = json.dumps(deployment, indent=2)
# # print(json_formatted_str)

# deployment, ip_hosts  = nested_deployment("toplevel", "foo1")
# json_formatted_str = json.dumps(deployment, indent=2)
# print(json_formatted_str)


# for ip_host in ip_hosts:
#     print(ip_host)
def parse_args(options):
    parser = argparse.ArgumentParser(
        description="Configs generator for nested deployment"
    )
    parser.add_argument(
        "-n",
        "--nested_role-distribution",
        type=str,
        default="nested_role_distribution.json",
    )
    parser.add_argument("--host", type=str, help="Hostname configuration target")
    parser.add_argument(
        "-d", "--deployment", type=str, help="Generated deploment name file"
    )
    parser.add_argument("-c", "--composition-info", type=str)
    parser.add_argument("--composition-name", type=str, default="composition")
    parser.add_argument(
        "-e",
        "--ip-hosts",
        type=str,
        help="Generated ip host tuple file (for /etc/hosts)",
    )
    parser.add_argument(
        "-i",
        "--ssh_pub_key_file",
        type=str,
        help="sshkey pub file (default ~/.ssh/id_rsa.pub)",
    )
    parser.add_argument(
        "-r", "--route-commands", type=str, help="Generated list off ip route commandes"
    )
    args = parser.parse_args(options)

    return args


# todo do we need host to nested table for ssh jump ?
def main(options=None):
    composition = {}

    inputs = parse_args(options)

    # read nested_role_
    role_distribution = read_role_distribution(inputs.nested_role_distribution)
    # import pdb; pdb.set_trace()
    #
    if inputs.composition_info:
        with open(realpath_from_store(inputs.composition_info), "r") as f:
            all_composition_info = json.load(f)
        composition = all_composition_info[inputs.composition_name]
    else:
        print("Composition info files is required")
        sys.exit(1)

    deployment_info, ip_hosts, route_commands, network = nested_deployment(
        composition, role_distribution, inputs.host
    )

    deployment = {
        "ssh_key.pub": get_ssh_pub_key(inputs.ssh_pub_key_file),
        "deployment": deployment_info,
        "composition": inputs.composition_name,
        "user": os.environ["USER"],
        "nested": True,
        "network": network,
    }

    if inputs.deployment:
        json_deployment = json.dumps(deployment, indent=2)
        with open(inputs.deployment, "w") as outfile:
            outfile.write(json_deployment)

    if inputs.ip_hosts:
        with open(inputs.ip_hosts, "w") as outfile:
            for ip_host in ip_hosts:
                outfile.write(ip_host + "\n")

    if inputs.route_commands:
        with open(inputs.route_commands, "w") as outfile:
            for route_command in route_commands:
                outfile.write(route_command + "\n")


if __name__ == "__main__":
    main()



import os
import os.path as op
import json
import subprocess
import click
import copy

from ..flavour import Flavour
from ..actions import read_compose_info, realpath_prefix_from_store
from ..driver.logger import rootlog
from ..driver.machine import Machine
from ..default_role import DefaultRole

from typing import Tuple, Optional


def set_prefix_store_volumes(dc_json, prefix_store):
    for service in dc_json["services"].keys():
        volumes = dc_json["services"][service]["volumes"]
        volumes_out = []
        for vol in volumes:
            if "/nix/store" == vol[:10]:
                volumes_out.append(f"{prefix_store}{vol[4:]}")
            else:
                volumes_out.append(vol)
        dc_json["services"][service]["volumes"] = volumes_out


def generate_docker_compose_file(ctx):
    base_docker_compose, prefix_store = realpath_prefix_from_store(
        ctx, ctx.compose_info["docker-compose-file"]
    )
    docker_compose_content = {"services": {}}
    nodes_info = {}

    artifact_dir = op.join(
        ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
    )
    os.makedirs(artifact_dir, mode=0o700, exist_ok=True)
    with open(base_docker_compose) as dc_file:
        dc_json = json.load(dc_file)
        if prefix_store:
            set_prefix_store_volumes(dc_json, prefix_store)

        roles_distribution = {}
        for role in ctx.compose_info["roles"]:
            if role in ctx.roles_distribution:
                roles_distribution[role] = ctx.roles_distribution[role]
            elif (
                "roles_distribution" in ctx.compose_info
                and role in ctx.compose_info["roles_distribution"]
            ):
                roles_distribution[role] = ctx.compose_info["roles_distribution"][role]
            else:
                roles_distribution[role] = 1

        # Add bind  deployment file inside containers
        deployment_file = op.join(
            ctx.envdir, f"deploy/{ctx.composition_flavour_prefix}.json"
        )

        deployment_symlink = op.join(artifact_dir, "deployment.json")
        if not os.path.islink(deployment_symlink):
            os.symlink(deployment_file, deployment_symlink)

        for service in dc_json["services"].values():
            (service["volumes"]).append(
                {
                    "type": "bind",
                    "source": deployment_symlink,
                    "target": "/etc/nxc/deployment.json",
                }
            )
        for role, distribution in roles_distribution.items():
            if type(distribution) is int:
                if distribution == 1:
                    hostname = f"{role}"
                    config = copy.copy(dc_json["services"][role])
                    config["hostname"] = hostname
                    docker_compose_content["services"][hostname] = config
                    nodes_info[hostname] = role
                else:
                    for i in range(1, distribution + 1):
                        hostname = f"{role}{i}"
                        config = copy.copy(dc_json["services"][role])
                        config["hostname"] = hostname
                        docker_compose_content["services"][hostname] = config
                        nodes_info[hostname] = role
            elif type(distribution) is list:
                for hostname in distribution:
                    config = copy.copy(dc_json["services"][role])
                    config["hostname"] = hostname
                    docker_compose_content["services"][hostname] = config
                    nodes_info[hostname] = role
            elif type(distribution) is DefaultRole:
                nb_min_nodes = distribution.nb_min_nodes
                ctx.log(
                    f"Docker: Using DefaultRole -> {nb_min_nodes} nodes for role '{role}'"
                )
                if nb_min_nodes == 1:
                    hostname = f"{role}"
                    config = copy.copy(dc_json["services"][role])
                    config["hostname"] = hostname
                    docker_compose_content["services"][hostname] = config
                    nodes_info[hostname] = role
                else:
                    for i in range(1, nb_min_nodes + 1):
                        hostname = f"{role}{i}"
                        config = copy.copy(dc_json["services"][role])
                        config["hostname"] = hostname
                        docker_compose_content["services"][hostname] = config
                        nodes_info[hostname] = role
            else:
                raise Exception("Unvalid type for specifying the roles of the nodes")
        dc_json.pop("services")
        docker_compose_content = docker_compose_content | dc_json

    artifact_dir = op.join(
        ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
    )
    os.makedirs(artifact_dir, mode=0o700, exist_ok=True)

    docker_compose_path = op.join(artifact_dir, "docker-compose.json")

    with open(docker_compose_path, "w") as outfile:
        outfile.write(json.dumps(docker_compose_content))
    return docker_compose_path, nodes_info


def generate_deployment_info_docker(ctx):
    if not ctx.compose_info:
        read_compose_info(ctx)

    deploy_dir = op.join(ctx.envdir, "deploy")
    if not op.exists(deploy_dir):
        create = click.style("   create", fg="green")
        ctx.log("   " + create + "  " + deploy_dir)
        os.mkdir(deploy_dir)

    docker_compose_path, nodes_info = generate_docker_compose_file(ctx)

    deployment_info = ctx.deployment_info

    # "nodes": ctx.compose_info["nodes"],
    deployment_info["nodes"] = list(nodes_info.keys())
    # "deployment": {n: {"role": n} for n in ctx.compose_info["nodes"]},
    deployment_info["deployment"] = {
        node_name: {"role": role_name} for (node_name, role_name) in nodes_info.items()
    }
    deployment_info["docker-compose-file"] = docker_compose_path

    if "test_script" in ctx.compose_info:
        deployment_info["test_script"] = ctx.compose_info["test_script"]

    if "all" in ctx.compose_info:
        deployment_info["all"] = ctx.compose_info["all"]

    # TODO move to action.py and factorize w/ geenerate_deployment_info
    with open(
        op.join(deploy_dir, f"{ctx.composition_flavour_prefix}.json"), "w"
    ) as outfile:
        outfile.write(json.dumps(deployment_info, indent=2))

    return docker_compose_path


class DockerFlavour(Flavour):
    docker_compose_file = None

    def __init__(self, ctx):
        super().__init__(ctx)

        self.name = "docker"
        self.description = ""
        # TOR self.docker_processes = {}

    def generate_deployment_info(self, ssh_pub_key_file):
        self.docker_compose_file = generate_deployment_info_docker(self.ctx)

    def driver_initialize(self, tmp_dir):
        assert self.ctx.deployment_info
        if not self.docker_compose_file:
            self.docker_compose_file = self.ctx.deployment_info["docker-compose-file"]

        nodes_names = self.ctx.deployment_info["nodes"]
        for name in nodes_names:
            self.machines.append(
                Machine(
                    self.ctx,
                    tmp_dir=tmp_dir,
                    start_command="",
                    name=name,
                )
            )

    def check(self, state="running"):
        check_process = subprocess.check_output(
            [
                "docker",
                "compose",
                "-f",
                self.docker_compose_file,
                "ps",
                "--services",
                "--filter",
                f"status={state}",
            ],
        )
        return len(check_process.decode().rstrip("\n").splitlines())

    def connect(self, machine):
        if machine.connected:
            return
        self.start_all()

    def start_all(self):
        if not self.external_connect:
            with rootlog.nested("starting docker compose"):
                subprocess.Popen(
                    ["docker", "compose", "-f", self.docker_compose_file, "up", "-d"]
                )

            self.wait_on_check()

        for machine in self.machines:
            if not machine.connected:
                self.start(machine)
                machine.connected = True

    def start(self, machine):  # TODO MOVE to Connect ???
        assert machine.name
        assert self.docker_compose_file

        machine.start_process_shell(
            [
                "docker",
                "compose",
                "-f",
                self.docker_compose_file,
                "exec",
                "-u",
                "root",
                "-T",
                machine.name,
                "bash",
                "-l",
            ]
        )

    def execute(
        self,
        machine,
        command: str,
        check_return: bool = True,
        timeout: Optional[int] = 900,
    ) -> Tuple[int, str]:
        return machine.execute_process_shell(command, check_return, timeout)

    def restart(self, machine):
        machine.restart_process_shell()

    def cleanup(self):
        # TODO handle stdout/stderr
        if not self.docker_compose_file:
            self.docker_compose_file = self.ctx.deployment_info["docker-compose-file"]
        subprocess.Popen(
            [
                "docker",
                "compose",
                "-f",
                self.docker_compose_file,
                "down",
                "--remove-orphans",
            ]
        )

    def shell_interact(self, machine) -> None:
        self.connect(machine)
        self.ext_connect("root", machine.name)

    def ext_connect(self, user, node, execute=True, ssh_key_file=None):
        if not self.docker_compose_file:
            self.docker_compose_file = self.ctx.deployment_info["docker-compose-file"]

        cmd = f"docker compose -f {self.docker_compose_file} exec -u {user} {node} bash"
        print(f"ext_connect {cmd}")
        if execute:
            return_code = subprocess.run(cmd, shell=True).returncode

            if return_code:
                self.ctx.wlog(f"Docker exit code is not null: {return_code}")
            return return_code
        else:
            return cmd



import click

import time
import os

import os.path as op

import sys
import glob

import ast
import json

import ptpython.repl

from ..context import pass_context, on_finished, on_started
from ..flavours import get_flavour_by_name

from ..actions import (
    read_deployment_info,
    read_test_script,
    push_on_machines,
    realpath_from_store,
)

from ..driver.driver import Driver
from ..httpd import HTTPDaemon
from ..setup import apply_setup


def start(ctx, interactive, execute_test_script, port, push_path=None):
    if (  # TODO rework (ask flavour ?)
        ctx.ip_addresses
        and (ctx.flavour.name != "vm-ramdisk")
        and (ctx.flavour.name != "vm")
    ) or ctx.flavour.name == "nspawn":
        if ctx.use_httpd:
            ctx.vlog("Launch: httpd to distribute deployment.json")
            ctx.httpd = HTTPDaemon(ctx=ctx, port=port)

        if hasattr(ctx.flavour, "generate_kexec_scripts"):
            ctx.flavour.generate_kexec_scripts()

        if push_path:
            push_on_machines(ctx)

        if ctx.use_httpd:
            ctx.httpd.start(directory=ctx.envdir)

        if not interactive:
            ctx.flavour.launch()
            sys.exit(0)

    test_script = read_test_script(ctx, ctx.compose_info)

    if not interactive and not execute_test_script:
        test_script = "start_all()"

    with Driver(
        # args.start_scripts, args.vlans, args.testscript.read_text(), args.keep_vm_state
        ctx,
        [],
        [],
        test_script,
        False,
    ) as driver:
        if interactive:
            ptpython.repl.embed(driver.test_symbols(), {})
        elif execute_test_script:
            tic = time.time()
            driver.run_tests()
            toc = time.time()
            ctx.glog(f"test script finished in {(toc-tic):.2f}s")
        else:
            ctx.glog("just start ???")
            driver.test_script()
    if ctx.use_httpd:
        ctx.httpd.stop()

    ctx.glog("Started")


# TODO define scope of dry_run, must you dry_run deployment file creation ?
# Also how to address driver part
# def dry_run_print(ctx, launch_cmd):
#     ctx.log("Dry-run:")
#     ctx.log("   launch command:              {launch_cmd}")
@click.command("start")
@click.option(
    "-I",
    "--interactive",
    is_flag=True,
    help="drop into a python repl with driver functions",
)
@click.option(
    "-m",
    "--machine-file",
    type=click.Path(resolve_path=True),
    help="file that contains remote machines names to (duplicates are considered as one).",
)
@click.option(
    "-W", "--wait-machine-file", is_flag=True, help="wait machine-file creation"
)
@click.option(
    "-s",
    "--ssh",
    type=click.STRING,
    default="ssh -l root ",
    help="specify particular ssh command",
)
@click.option(
    "-S",
    "--sudo",
    type=click.STRING,
    default="sudo",
    help="specify particular sudo command",
)
@click.option(
    "--push-path",
    help="remote path where to push image, kernel and kexec_script on machines (use to re-kexec)",
)
@click.option(
    "--reuse",
    is_flag=True,
    help="supposed a previous succeded start (w/ root access via ssh)",
)
@click.option(
    "--remote-deployment-info",
    "--http-deployment-info",
    is_flag=True,
    help="deployment info is served via http (in place of kernel parameters)",
)
@click.option(
    "--port",
    type=click.INT,
    default=0,
    help="Port to use for the HTTP server",
)
@click.option(
    "--image-store-ssh",
    type=click.STRING,
    help="(experimental, only for Grid'5000) [username@]hostname[:port]",
)
@click.option(
    "-c",
    "-C",
    "--composition",
    type=click.STRING,
    help="specify composition, can specify flavour e.g. composition::flavour",
)
@click.option(
    "-f",
    "--flavour",
    type=click.STRING,
    help="specify flavour",
)
@click.option(
    "-t",
    "--test-script",
    is_flag=True,
    help="execute testscript",
)
@click.option(
    "--file-test-script",
    type=click.STRING,
    help="alternative testscript",
)
@click.option(
    "-w",
    "--sigwait",
    is_flag=True,
    help="wait any signal to exit after a start only action (not testscript execution or interactive use",
)
@click.option(
    "-k",
    "--kernel-params",
    type=click.STRING,
    help="additional kernel parameters, this option is flavour dependent",
)
@click.option(
    "-r",
    "--role-distribution",
    multiple=True,
    help="specify the number of nodes or nodes' name for a role (e.g. compute=2 or server=foo,bar ).",
)
@click.argument(
    "roles_distribution_file",
    required=False,
    default=None,
    type=click.Path(exists=True),
)
@click.option(
    "--compose-info",
    type=click.Path(resolve_path=True),
    help="specific compose info file",
)
@click.option(
    "-i",
    "--identity-file",
    type=click.STRING,
    help="path to the ssh public key to use to connect to the deployments",
)
@click.option(
    "-s",
    "--setup",
    type=click.STRING,
    help="Select setup variant",
)
@click.option(
    "-p",
    "--parameter",
    type=click.STRING,
    multiple=True,
    help="Parameter added to deployment file (for contextualization phase)",
)
@click.option(
    "-P",
    "--parameter-file",
    type=click.STRING,
    help="Json file contains parameters added to deployment file (for contextualization phase)",
)
@click.option(
    "-d",
    "--deployment-file",
    type=click.STRING,
    help="Deployement json file use for the deployment (skip generation) Warning parametrization not supported (upto now)",
)
@click.option(
    "--ip-range",
    type=click.STRING,
    default="",
    help="IP range (for now only usable with nspawn flavour)",
)
@click.option(
    "-a",
    "--tag",
    type=click.STRING,
    help="tagname of the deployment",
)
@click.option(
    "--start-option",
    type=click.STRING,
    multiple=True,
    help="Additional option to modify start process (flavour dependant, see documentation)",
)
#     "--dry-run", is_flagmultiple=True,=True, help="Show what this command would do without doing it"
# )
@pass_context
@on_finished(lambda ctx: ctx.show_elapsed_time())
@on_started(lambda ctx: ctx.warning_valid_env())
def cli(
    ctx,
    interactive,
    machine_file,
    wait_machine_file,
    ssh,
    sudo,
    push_path,
    reuse,
    composition,
    flavour,
    remote_deployment_info,
    port,
    test_script,
    file_test_script,
    sigwait,
    kernel_params,
    role_distribution,
    roles_distribution_file,
    compose_info,
    identity_file,
    setup,
    parameter,
    parameter_file,
    ip_range,
    deployment_file,
    tag,
    image_store_ssh,
    start_option
    # dry_run,
):
    """
    Starts a set of nodes using the previous build.

    `ROLE_DISTRIBUTION_FILE` is and optional YAML file describing how many instances of each role are expected.

    ## Examples

    - `nxc start`

       Start the last built composition.

    - `nxc start role-distrib.yaml`

        With the file `role-distrib.yaml` written as this:

        ```yaml
        nfsServerNode: 1
        nfsClientNode: 2
        ```

        Instantiates two nodes with the role `nfsClientNode` and one only with the role `nfsServerNode`. Of course, these roles have to be described beforehand in a `composition.nix` file.

    """
    flavour_name = flavour
    if test_script:
        execute_test_script = True
        test_script = None
    else:
        execute_test_script = False

    if file_test_script:
        raise click.ClickException(
            "alternative test-scipt execution not yet implemented"
        )

    ctx.log("Starting")
    ctx.ssh = ssh
    ctx.sudo = sudo
    ctx.interactive = interactive
    ctx.execute_test_script = execute_test_script
    ctx.sigwait = sigwait
    ctx.ip_range = ip_range
    ctx.machine_file = machine_file
    ctx.image_store_ssh = image_store_ssh
    ctx.start_option = start_option
    if deployment_file:
        if not flavour:
            ctx.elog("Option --flavour is required with --deployment-file option !")
            sys.exit(2)
        else:
            ctx.flavour = get_flavour_by_name(flavour_name)(ctx)
            read_deployment_info(ctx, deployment_file)

    if deployment_file and (role_distribution or roles_distribution_file):
        ctx.wlog(
            "--role-distribution and --roles-distribution-file are ignored with --deployment-file option !"
        )

    ctx.set_roles_distribution(role_distribution, roles_distribution_file)

    # kernel_params can by set through setup
    if setup or op.exists(op.join(ctx.envdir, "setup.toml")):
        _, _, _, _, kernel_params = apply_setup(
            ctx,
            setup,
            None,
            None,
            None,
            None,
            None,
            kernel_params,
        )
    ctx.kernel_params = kernel_params

    if remote_deployment_info:
        ctx.use_httpd = True

    if parameter_file:
        with open(parameter_file, "r") as f:
            try:
                ctx.deployment_info["parameters"] = json.load(f)
            except ValueError:
                raise click.ClickException(
                    f"Failed to parse parameters json file: {parameter_file}"
                )

    if parameter:
        if "parameters" not in ctx.deployment_info:
            ctx.deployment_info["parameters"] = {}
        for p in parameter:
            try:
                (k, v) = p.split("=", 1)
                v = ast.literal_eval(v)
            except ValueError:
                raise click.ClickException(f"Fail to parse parameter: {p}")
            ctx.deployment_info["parameters"][k] = v

    if deployment_file:
        start(ctx, interactive, execute_test_script, port, push_path)
        sys.exit(0)

    build_path = op.join(ctx.envdir, "build")

    if not compose_info and not op.exists(build_path):
        raise click.ClickException(
            "You need build composition first, with nxc build command"
        )

    # Handle cases where machines list must be provided
    if machine_file and not op.isfile(machine_file) and not wait_machine_file:
        raise click.ClickException(f"{machine_file} file does not exist")

    if push_path and not machine_file:
        raise click.ClickException("machine_file must be provided to use push_path")

    if wait_machine_file:
        if not machine_file:
            raise click.ClickException(
                "You need to provide --machine-file option with --wait"
            )

        if not op.isfile(machine_file):
            # ctx.log(f"Waiting {machine_file} file creation")
            # TODO: add quiet option
            if ctx.show_spinner:
                ctx.spinner.start(f"Waiting for {machine_file} creation")

            # Note: inotify approach does not work with NFS, and pyinotify
            # is no more developed so simple pulling loop is used
            # if "nfs" == get_fs_type(machine_file):
            while not op.isfile(machine_file):
                time.sleep(0.1)

            if ctx.show_spinner:
                ctx.spinner.succeed(f"{machine_file} file created")
            else:
                ctx.log(f"{machine_file} file created")

    # Determine composition and flavour name
    # First case composition is given not flavour name
    if composition and (flavour_name is None):
        splitted_composition = composition.split("::")
        len_splitted_composition = len(splitted_composition)
        if len_splitted_composition == 2:
            composition_name, flavour_name = splitted_composition
            composition_all_in_one_file = op.join(ctx.envdir, f"build/::{flavour_name}")

            if not op.lexists(composition_all_in_one_file):
                build_path = op.join(ctx.envdir, f"build/{composition}")
                if not op.lexists(build_path):
                    raise click.ClickException(
                        f"Build file does not exist: {build_path}"
                    )
            else:
                build_path = composition_all_in_one_file

            ctx.flavour = get_flavour_by_name(flavour_name)(ctx)
            ctx.composition_name = composition_name
            ctx.composition_flavour_prefix = (
                (composition + "::" + tag) if tag else composition
            )
            ctx.composition_basename_file = composition_name
        else:
            raise Exception(
                "Sorry, provide only flavour or only composition is not supported"
            )

    if composition is None and compose_info is None:
        if flavour_name:
            search_path = f"{build_path}/*::{flavour_name}"
        else:
            search_path = f"{build_path}/*"

        build_paths = glob.glob(search_path)
        if not build_paths:
            raise click.ClickException("Failed to find last build")

        last_build_path = max(
            build_paths,
            key=lambda x: os.stat(x, follow_symlinks=False).st_ctime,
        )

        ctx.log("Use last build:")
        ctx.glog(last_build_path)

        build_path = last_build_path
        ctx.composition_flavour_prefix = (
            (op.basename(last_build_path) + "::" + tag)
            if tag
            else str(op.basename(last_build_path))
        )

        splitted_basename = ctx.composition_flavour_prefix.split("::")

        if splitted_basename[0] == "":
            raise click.ClickException("Sorry, composition name must be provided")

        ctx.composition_name = splitted_basename[0]
        ctx.composition_basename_file = ctx.composition_name

        if flavour_name:
            assert flavour_name == splitted_basename[1]

        ctx.flavour = get_flavour_by_name(splitted_basename[1])(ctx)

    if not compose_info and op.isdir(build_path) and len(os.listdir(build_path)) == 0:
        ctx.wlog(f"{build_path} is an empty directory, surely a nixos-test result !")
        sys.exit(2)

    if (composition is None) and flavour_name and compose_info:
        ctx.flavour = get_flavour_by_name(flavour_name)(ctx)
        ctx.compose_info_file = realpath_from_store(ctx, compose_info)
        ctx.composition_flavour_prefix = (
            (op.basename(compose_info) + "::" + tag)
            if tag
            else op.basename(compose_info)
        )
        ctx.composition_name = "composition"
        ctx.composition_basename_file = ctx.composition_name
    else:
        ctx.compose_info_file = realpath_from_store(ctx, build_path)

    #
    # ssh not used, subsequent_start_values and first_start_values not use w/
    # nfs mount on g5k. If not need on other platform to REMOVE (see also platform.py
    # Only depends of flavours or plaforms
    #
    # if ctx.platform:
    #     if reuse:
    #         (ssh, sudo, push_path) = ctx.platform.subsequent_start_values
    #     else:
    #         (ssh, sudo, push_path) = ctx.platform.first_start_values

    ctx.flavour.generate_deployment_info(identity_file)
    start(ctx, interactive, execute_test_script, port, push_path)



#!/usr/bin/env python3
import sys
import argparse
import asyncio
from string import Template

CMD_BASE = Template(
    r'echo "s=\$$(mktemp); echo \\"$inner_cmd\\" > \$$s; nohup bash \$$s > /dev/null 2> /dev/null < /dev/null &" | $ssh $host "bash -s"'
)

CMD_START = Template(
    "until nc -z $host $port1 ; do sleep 0.01; done; base64 $file_input  >/dev/tcp/$host/$port0"
)
CMD_END = Template("nc -l $port0 | base64 -d > $file_output & nc -l $port1")
CMD_TEE = Template(
    "until nc -z $host $port3; do sleep 0.01; done; nc -l $port0 | tee >(cat > /dev/tcp/$host/$port2) | base64 -d > $file_output & nc -l $port1"
)

# B  = cmd_base.substitute({'inner_cmd' : 'until nc -z localhost 5556; do sleep 0.01; done; nc -l 4444 | tee >(cat > /dev/tcp/127.0.0.1/5555) | base64 -d > /tmp/yopB & nc -l 4446'})

# A = 'until nc -z localhost 4446; do sleep 0.01; done; base64 /tmp/vm-state-client1/client1.qcow2  >/dev/tcp/127.0.0.1/4444'


def elog(msg, *args):
    print("\033[91mError:\033[0m", *args, file=sys.stderr)


def vlog(msg, *args):
    print(*args)


async def run(cmd):
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        executable="/bin/bash",
    )

    stdout, stderr = await proc.communicate()

    return (proc.returncode, stdout, stderr)


def generate_pipe_tasks(
    hosts, file_input, file_output, port0="5555", port1="5556", ssh="ssh"
):
    hosts_rev = hosts.copy()
    hosts_rev.reverse()

    def cmd_tee(h, h_next):
        # print(h, h_next)
        cmd_tee = CMD_TEE.substitute(
            {
                "host": h_next,
                "port1": port1,
                "port0": port0,
                "port3": port1,
                "port2": port0,
                "file_output": file_output,
            }
        )
        return CMD_BASE.substitute({"inner_cmd": cmd_tee, "ssh": ssh, "host": h})

    tees = [cmd_tee(h, hosts_rev[i]) for i, h in enumerate(hosts_rev[1:])]

    cmd_end = CMD_END.substitute(
        {"port1": port1, "port0": port0, "file_output": file_output}
    )
    end = CMD_BASE.substitute({"inner_cmd": cmd_end, "ssh": ssh, "host": hosts_rev[0]})

    start = CMD_START.substitute(
        {"host": hosts[0], "port1": port1, "port0": port0, "file_input": file_input}
    )

    tasks_cmd = [end] + tees + [start]

    # for c in tasks_cmd:
    #    print(c)
    #    print("")

    return tasks_cmd


def generate_scp_tasks(hosts, file_input, file_output, scp="scp", user=""):
    if user:
        user = f"{user}@"
    tasks_cmd = [f"{scp} {file_input} {user}{h}:{file_output}" for h in hosts]
    # print(tasks_cmd)
    return tasks_cmd


def exec_kataract_tasks(tasks_cmd, elog=elog, vlog=vlog):
    asyncio.set_event_loop(asyncio.new_event_loop())
    loop = asyncio.get_event_loop()

    tasks = [run(task_cmd) for task_cmd in tasks_cmd]

    finished, unfinished = loop.run_until_complete(asyncio.wait(tasks, timeout=60))

    if len(unfinished) > 0:
        elog("Failed")

    vlog(f"unfinished: {len(unfinished)}")
    vlog(f"finished: {len(finished)}")

    for task in finished:
        r = task.result()
        if r[0] != 0 or r[2]:
            vlog(f"{r}")
    loop.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="kataract",
        description="File broadcaster based on TCP Pipeline. Pipeline is initiated by bash command sent through ssh connection",
    )
    parser.add_argument(
        "--input", "-i", dest="file_input", required=True, help="input file to send"
    )
    parser.add_argument(
        "--output",
        "-o",
        dest="file_output",
        required=True,
        help="output file to write on receiver hosts",
    )
    parser.add_argument(
        "--ssh",
        "-s",
        dest="ssh",
        default="ssh",
        help="ssh command to use to reach host, 'ssh' by default",
    )
    parser.add_argument(
        "-m", dest="hosts", action="append", required=True, help="receiver hosts"
    )
    parser.add_argument(
        "--port-data",
        "-p",
        dest="port_data",
        default=5555,
        type=int,
        help="port to send data",
    )
    parser.add_argument(
        "--port-ready",
        "-r",
        dest="port_ready",
        default=5556,
        type=int,
        help="port to test if host is ready to receive.",
    )
    parser.add_argument(
        "--scp", "-S", dest="scp", action="store_true", help="use scp to send file."
    )

    # main()
    args = parser.parse_args()

    # if not args.file_input()
    if not args.scp:
        tasks_cmd = generate_pipe_tasks(
            args.hosts,
            args.file_input,
            args.file_output,
            port0=args.port_data,
            port1=args.port_ready,
            ssh=args.ssh,
        )
    else:
        tasks_cmd = generate_scp_tasks(
            args.hosts, args.file_input, args.file_output, scp="scp"
        )

    exec_kataract_tasks(tasks_cmd)



import os

# import sys
# import json
# import base64
# from ..httpd import HTTPDaemon

from ..flavour import Flavour
from ..actions import (
    generate_deployment_info,
    ssh_connect,
    kill_proc_tree,
    realpath_from_store,
)
from ..driver.vlan import VLan
from ..driver.logger import rootlog
from ..driver.machine import Machine, StartScript
from ..platform import platform_detection


class VmBasedFlavour(Flavour):

    """
    The Vm Ramdisk flavour. This is flavour provides a system image to be executed with QEMU and use memory only for root system. By consequence lot of ram is used around 2Go minimum by node.
    """

    vm = True
    tmp_dir = None
    vlan = None

    def __init__(self, ctx):
        super().__init__(ctx)
        ctx.external_connect = True  # to force use of ssh on foo.execute(command)
        platform_detection(ctx)

    def generate_deployment_info(self, ssh_pub_key_file=None):
        generate_deployment_info(self.ctx, ssh_pub_key_file)

    def create_machines(self):
        ctx = self.ctx
        deployment_nodes = ctx.deployment_info["deployment"]
        qemu_script = realpath_from_store(
            ctx, ctx.deployment_info["all"]["qemu_script"]
        )

        for node in deployment_nodes.values():
            start_command = ""
            if not ctx.no_start:
                start_command = StartScript(qemu_script, node["vm_id"], self)
            self.machines.append(
                Machine(
                    ctx,
                    ip="127.0.0.1",
                    ssh_port=f"{22021 + int(node['vm_id'])}",
                    tmp_dir=self.tmp_dir,
                    start_command=start_command,
                    keep_vm_state=False,
                    name=node["host"],
                )
            )

    def driver_initialize(self, tmp_dir):
        self.tmp_dir = tmp_dir
        ctx = self.ctx
        # deployment = ctx.deployment_info

        if ctx.no_start:
            self.create_machines()
            # deployment_nodes = ctx.deployment_info["deployment"]
            # for ip, node in deployment_nodes.items():
            #     ssh_port = 22
            #     if "ssh-port" in node:
            #         ssh_port = node["ssh-port"]
            #     self.machines.append(
            #         Machine(
            #             ctx,
            #             ip=ip,
            #             ssh_port = ssh_port,
            #             tmp_dir=tmp_dir,
            #             start_command="",(qemu_script, v["vm_id"], self)
            #             keep_vm_state=False,
            #             name=node["host"],
            #         )
            #     )

            for machine in self.machines:
                if not machine.connected:
                    self.start(machine)
                machine.connected = True
            return

        # os.environ["KERNEL"] = ctx.deployment["all"]["kernel"]
        # os.environ["INITRD"] = ctx.deployment["all"]["initrd"]
        # base_qemu_script = ctx.deployment["all"]["qemu_script"]

        # debug_stage1 = None
        # debug_var_base = ""
        # if "DEBUG_STAGE1" in os.environ:
        #     debug_stage1 = os.environ["DEBUG_STAGE1"]

        #########
        # Determine DEPLOY data
        #########
        # if "DEPLOY" not in os.environ:
        #     if ctx.use_httpd:
        #         if not ctx.httpd:
        #             ctx.httpd = HTTPDaemon(ctx=ctx)
        #             ctx.httpd.start(directory=ctx.envdir)
        #         base_url = f"http://{ctx.httpd.ip}:{ctx.httpd.port}"
        #         deploy_info_src = (
        #             f"{base_url}/deploy/{ctx.composition_flavour_prefix}.json"
        #         )
        #     else:
        #         if not ctx.deployment_info_b64:
        #             deployment_info_str = json.dumps(deployment)
        #             deploy_info_src = base64.b64encode(
        #                 deployment_info_str.encode()
        #             ).decode()
        #         else:
        #             deploy_info_src = ctx.deployment_info_b64

        #         if len(deploy_info_src) > (4096 - 256):  # 4096 -> 2048 ???
        #             rootlog.nested(
        #                 "The base64 encoded deploy data is too large: use an http server to serve it"
        #             )
        #             sys.exit(1)
        #     os.environ["DEPLOY"] = f"deploy={deploy_info_src}"
        #     print(f"deploy={deploy_info_src}")
        # else:
        #     rootlog.nested(f'Variable environment DEPLOY: {os.environ["DEPLOY"]}')
        if ctx.platform and ctx.platform.nix_store:
            os.environ["SHARED_NIX_STORE_DIR"] = ctx.platform.nix_store
            os.environ["KERNEL"] = realpath_from_store(
                ctx, ctx.deployment_info["all"]["kernel"]
            )
            os.environ["INITRD"] = realpath_from_store(
                ctx, ctx.deployment_info["all"]["initrd"]
            )
            ctx.vlog(f"KERNEL: {os.environ['KERNEL']}")
            ctx.vlog(f"INITRD: {os.environ['INITRD']}")

        if "DEPLOY" not in os.environ:
            os.environ[
                "DEPLOY"
            ] = f"deploy={self.ctx.deployment_filename[len(self.ctx.envdir)+1:]}"
        else:
            rootlog.nested(f'Variable environment DEPLOY: {os.environ["DEPLOY"]}')

        self.create_machines()

    def create_vlan(self):
        self.vlan = VLan(0, self.tmp_dir, ctx=self.ctx)
        return self.vlan

    def start_process_shell(self, machine):
        machine.start_process_shell(
            [
                "ssh",
                "-t",
                "-o",
                "StrictHostKeyChecking=no",
                "-l",
                "root",
                "-p",
                machine.ssh_port,
                machine.ip,
            ]
        )

    def start(self, machine):
        if not self.ctx.no_start:
            machine._start_vm()
        else:
            self.start_process_shell(machine)

    def release(self, machine):
        if machine.pid is None:
            return
        rootlog.info(f"kill machine (pid {machine.pid})")
        assert machine.process
        assert machine.shell
        assert machine.monitor
        assert machine.serial_thread

        # Kill children
        kill_proc_tree(machine.pid, include_parent=False)

        machine.process.terminate()
        machine.shell.close()
        machine.monitor.close()
        machine.serial_thread.join()

    def ext_connect(self, user, node, execute=True, ssh_key_file=None):
        return ssh_connect(self.ctx, user, node, execute, ssh_key_file)


class VmFlavour(VmBasedFlavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.name = "vm"
        self.image = {"distribution": "all-in-one"}
        self.description = "VM with shared nix-store and stage-1"


class VmRamdiskFlavour(VmBasedFlavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.name = "vm-ramdisk"
        self.image = {"type": "ramdisk", "distribution": "all-in-one"}
        self.description = "Plain vm ramdisk (all-in-memory), need lot of ram !"



import os
import os.path as op
import importlib.metadata

import click

from .context import pass_context, CONTEXT_SETTINGS

click.disable_unicode_literals_warning = True
version = importlib.metadata.version("nixos-compose")


class NixosComposeCLI(click.MultiCommand):
    def list_commands(self, ctx):
        cmd_folder = op.abspath(op.join(op.dirname(__file__), "commands"))
        commands = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith(".py") and filename.startswith("cmd_"):
                commands.append(filename[4:-3])
        commands.sort()
        return commands

    def get_command(self, ctx, name):
        if name in self.list_commands(ctx):
            mod = __import__("nixos_compose.commands.cmd_" + name, None, None, ["cli"])
            return mod.cli


@click.command(cls=NixosComposeCLI, context_settings=CONTEXT_SETTINGS, chain=True)
@click.option(
    "--envdir",
    "-d",
    type=click.Path(file_okay=False, resolve_path=True),
    default=op.abspath("./nxc"),
    help="Changes the folder to operate on.",
)
@click.option("--verbose", "-v", is_flag=True, default=False, help="Verbose mode.")
@click.option("--debug", "-D", is_flag=True, default=False, help="Enable debugging")
@click.version_option(version=version)
@pass_context
def nxc(ctx, envdir, verbose, debug):
    """Generate and manage Nixos-compositions."""
    ctx.envdir = envdir
    if os.path.isfile("nxc.json"):
        if os.path.islink("nxc.json"):
            ctx.nxc_file = op.abspath(os.readlink("nxc.json"))
        else:
            ctx.nxc_file = op.abspath("nxc.json")
        with open(ctx.nxc_file, "r") as f:
            ctx.load_nxc(f)

        ctx.envdir = op.dirname(ctx.nxc_file)

    ctx.verbose = verbose
    ctx.debug = debug
    # ctx.update() not use


def main():
    nxc()



from execo import Host
from execo_g5k import get_oar_job_nodes

import os
import os.path as op
from pathlib import Path

import tempfile
from .context import Context
from .actions import realpath_from_store, translate_hosts2ip, wait_ssh_ports
from .flavours import get_flavour_by_name

# from .g5k import key_sleep_script
from .httpd import HTTPDaemon


def get_envdir(ctx):
    if os.path.isfile("nxc.json"):
        if os.path.islink("nxc.json"):
            ctx.nxc_file = os.readlink("nxc.json")
        else:
            ctx.nxc_file = op.abspath("nxc.json")
        with open(ctx.nxc_file, "r") as f:
            ctx.load_nxc(f)

        ctx.envdir = op.dirname(ctx.nxc_file)
    else:
        raise Exception("Cannot find `nxc.json`")


def get_oar_job_nodes_nxc(
    oar_job_id,
    site,
    compose_info_file=None,
    flavour_name="g5k-nfs-store",
    composition_name="composition",
    roles_quantities={},
    port=0,
    skip_deploy=False,
    image_store_ssh=None,
):
    """
    Brother of the "get_oar_job_nodes" function from execo
    but does the mapping with roles from NXC
    """
    ctx = Context()
    # TODO: kaberk
    ctx.composition_name = composition_name
    ctx.flavour_name = flavour_name
    ctx.composition_flavour_prefix = f"{composition_name}::{flavour_name}"
    ctx.roles_distribution = roles_quantities
    ctx.image_store_ssh = image_store_ssh
    ctx.envdir = None
    get_envdir(ctx)

    if compose_info_file:
        ctx.compose_info_file = compose_info_file
    else:
        build_folder = op.join(ctx.envdir, "build")
        simlink_build = op.join(
            build_folder, f"{ctx.composition_name}::{ctx.flavour_name}"
        )
        ctx.compose_info_file = realpath_from_store(ctx, simlink_build)

    flavour = get_flavour_by_name(flavour_name)(ctx)
    ctx.flavour = flavour

    # print(f"compose info file: {ctx.compose_info_file}")

    if "g5k" in flavour_name:
        g5k_nodes = get_oar_job_nodes(oar_job_id, site)
        # print(f"G5K nodes: {g5k_nodes}")
        machines = [node.address for node in g5k_nodes]
        translate_hosts2ip(ctx, machines)

        flavour.generate_deployment_info()
    elif flavour_name == "vm":
        flavour.generate_deployment_info()

        # TODO use a tempfile.TemporaryDirectory ?
        tmp_dir = Path(os.environ.get("TMPDIR", tempfile.gettempdir()))
        tmp_dir.mkdir(mode=0o700, exist_ok=True)
        flavour.driver_initialize(tmp_dir)
        # TODO _vlan = flavour.create_vlan()
        machines = list(map(lambda x: (x.ip, x.ssh_port), flavour.machines))
        flavour.ctx.ip_addresses = list(map(lambda x: x[0], machines))
    else:
        raise Exception(
            f"Unsupported flavour '{flavour}' for the Execo-NXC integration"
        )

    if len(machines) > 3:
        ctx.use_httpd = True
        ctx.httpd = HTTPDaemon(ctx=ctx, port=port)
        ctx.httpd.start(directory=ctx.envdir)

    if not skip_deploy:
        ctx.log("Deploying")
        if hasattr(flavour, "generate_kexec_scripts"):
            flavour.generate_kexec_scripts()
            flavour.launch()
        elif flavour_name == "vm":
            for machine in flavour.machines:
                machine.start()
            wait_ssh_ports(ctx)
        else:
            user = os.environ["USER"]
            tempfile.tempdir = f"/home/{user}/public"
            tmp = tempfile.NamedTemporaryFile(delete=False)
            tmp_kaenv = tempfile.NamedTemporaryFile(delete=False)
            temp_dir = tempfile.TemporaryDirectory()
            try:
                machines_str = "\n".join(f"{m}" for m in machines)
                tmp.write(machines_str.encode("utf-8"))
                tmp.flush()
                nxc_image_path = op.join(temp_dir.name, "nixos.tar.xz")
                flavour.ask_before_kadeploy = False
                ctx.machine_file = tmp.name
                flavour.launch(
                    machine_file=tmp.name,
                    kaenv_path=tmp_kaenv.name,
                    deploy_image_path=nxc_image_path,
                )
            finally:
                tmp.close()
                os.unlink(tmp.name)
                tmp_kaenv.close()
                os.unlink(tmp_kaenv.name)
                temp_dir.cleanup()

    roles = {}
    nodes = {}
    for ip_addr, node_info in flavour.ctx.deployment_info["deployment"].items():
        node_role = node_info["role"]
        port = 22 if flavour_name != "vm" else 22021 + int(node_info["vm_id"])
        ip_addr_node = ip_addr if flavour_name != "vm" else "localhost"
        localhost = Host(ip_addr_node, user="root", port=port)
        nodes[node_info["host"]] = localhost
        if node_role in roles:
            roles[node_role].append(localhost)
        else:
            roles[node_role] = [localhost]

    if ctx.use_httpd:
        ctx.httpd.stop()
    return (nodes, roles)



import os
import os.path as op
import sys
import time
from string import Template
import click
import subprocess
import socket
import json

from ..flavour import Flavour
from ..actions import (
    get_machine_from_file,
    read_compose_info,
    realpath_from_store_remote,
    generate_deployment_info,
    generate_deploy_info_b64,
    generate_kexec_scripts,
    launch_ssh_kexec,
    wait_ssh_ports,
    ssh_connect,
)
from ..driver.machine import Machine

# from ..driver.logger import rootlog

KADEPLOY_ARCH = {
    "x86_64-linux": "x86_64",
    "powerpc64le-linux": "ppc64le",
    "aarch64-linux": "aarch64",
}

KADEPLOY_ENV_DESC = """
      name: $image_name
      version: 1
      description: NixOS
      author: $author
      visibility: shared
      destructive: false
      os: linux
      arch: $system
      image:
        file: $file_image_url
        kind: tar
        compression: xz
      boot:
        kernel: /boot/bzImage
        initrd: /boot/initrd
        kernel_params: $kernel_params
      filesystem: ext4
      partition_type: 131
      multipart: false
"""


def generate_kadeploy_envfile(
    ctx, deploy=None, kernel_params="", kaenv_path=None, deploy_image_path=None
):
    if not ctx.compose_info:
        read_compose_info(ctx)

    base_path = op.join(
        ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
    )
    os.makedirs(base_path, mode=0o700, exist_ok=True)
    if kaenv_path is None:
        kaenv_path = op.join(base_path, "nixos.yaml")

    if not deploy:
        if ctx.use_httpd:
            base_url = f"http://{ctx.httpd.ip}:{ctx.httpd.port}"
            deploy = f"{base_url}/deploy/{ctx.composition_flavour_prefix}.json"
        else:
            generate_deploy_info_b64(ctx)
            deploy = ctx.deployment_info_b64

    user = os.environ["USER"]

    fqdn = socket.getfqdn()
    g5k_site = fqdn.split(".")[1]

    system = ctx.compositions_info["system"]
    additional_kernel_params = ""
    if ctx.kernel_params:
        additional_kernel_params = ctx.kernel_params
    with open(kaenv_path, "w") as kaenv_file:
        t = Template(KADEPLOY_ENV_DESC)
        kaenv = t.substitute(
            image_name="NixOS",
            author=user,
            system=KADEPLOY_ARCH[system],
            file_image_url=f"local://{deploy_image_path}"
            if deploy_image_path
            else f"http://public.{g5k_site}.grid5000.fr/~{user}/nixos.tar.xz",
            kernel_params=f"boot.shell_on_fail console=tty0 console=ttyS0,115200 check_kernel_initrd deploy={deploy} {additional_kernel_params} {kernel_params}",
        )
        kaenv_file.write(kaenv)


def generate_machine_file_retrieve_ips(ctx):
    if not ctx.machine_file:
        fqdn = socket.getfqdn()
        g5k_site = fqdn.split(".")[1]
        g5k_frontend = "f" + g5k_site
        if g5k_frontend != socket.gethostname():
            output = subprocess.check_output(
                ["ssh", g5k_frontend, "oarstat -u -J"]
            ).decode()
        else:
            output = subprocess.check_output(["oarstat", "-u", "-J"]).decode()
        oarstat_json = json.loads(output)

        job_id = 0
        nb_nodes = 0
        for jid, j in oarstat_json.items():
            if "deploy" in j["types"]:
                try:
                    os.remove(".oar_nodefile")
                except OSError:
                    pass
                with open(".oar_nodefile", "w") as outfile:
                    for n in j["assigned_network_address"]:
                        outfile.write(n + "\n")
                        nb_nodes += 1
                        job_id = jid

                if nb_nodes > 0:
                    ctx.vlog(
                        f"Auto generate .oar_nodefile as a machine file from job: {job_id} with {nb_nodes} nodes identified"
                    )
                else:
                    ctx.elog(
                        "Cannot retrieve machines from existing deployed job, verify it exists or give a machine file"
                    )
                    sys.exit(1)
                ctx.machine_file = ctx.envdir + "/.oar_nodefile"
                break
    if not ctx.machine_file:
        ctx.elog(
            "Cannot retrieve machines from any existing jobs, verify if one exists or give a machine file"
        )
        sys.exit(1)
    get_machine_from_file(ctx)


class G5kFlavour(Flavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        if ctx.ssh == "":
            ctx.ssh = "ssh -l root"

    def generate_deployment_info(self, ssh_pub_key_file=None):
        if self.ctx.ip_addresses == []:
            generate_machine_file_retrieve_ips(self.ctx)
        generate_deployment_info(self.ctx, ssh_pub_key_file)


class G5kKexecBasedFlavour(G5kFlavour):
    def __init__(self, ctx):
        super().__init__(ctx)

    def generate_kexec_scripts(self):
        generate_kexec_scripts(self.ctx)

    def launch(self):
        if "no-launch" in self.ctx.start_option:
            self.ctx.vlog("Start option no-launch, exit now without launching kexec")
            exit(0)
        launch_ssh_kexec(self.ctx)
        time.sleep(10)
        wait_ssh_ports(self.ctx)

    def driver_initialize(self, tmp_dir):
        self.tmp_dir = tmp_dir
        ctx = self.ctx

        if ctx.no_start:  #
            deployment_nodes = self.ctx.deployment_info["deployment"]
            for ip, node in deployment_nodes.items():
                self.machines.append(
                    Machine(
                        self.ctx,
                        ip=ip,
                        tmp_dir=tmp_dir,
                        start_command="",
                        keep_vm_state=False,
                        name=node["host"],
                    )
                )

            for machine in self.machines:
                if not machine.connected:
                    self.start(machine)
                machine.connected = True
            return

    def start(self, machine):
        if not self.ctx.no_start:
            print("Not Yet Implemented")
            exit(1)
        else:
            machine.start_process_shell(
                [
                    "ssh",
                    "-t",
                    "-o",
                    "StrictHostKeyChecking=no",
                    "-l",
                    "root",
                    machine.ip,
                ]
            )

    def ext_connect(self, user, node, execute, ssh_key_file):
        return ssh_connect(self.ctx, user, node, execute, ssh_key_file)


class G5kNfsStoreFlavour(G5kKexecBasedFlavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.name = "g5k-nfs-store"

    def generate_kexec_scripts(self):
        def get_store_path():
            for prefix_store_path in self.ctx.alternative_stores + ["/nix"]:
                store_path = f"{prefix_store_path}/store"
                if op.exists(store_path):
                    return store_path
            self.ctx.elog("Store Path Not Found")
            exit(1)

        if "NFS_STORE" in os.environ:
            nfs_store_str = os.environ["NFS_STORE"]
            nfs_store_parts = nfs_store_str.split(":")

            try:
                # use ip tp avoid potential issue during stage 1 on node if fqn is not used
                ip = socket.gethostbyname(nfs_store_parts[0])
            except ValueError:
                self.ctx.elog(
                    f"NFS_STORE environment variable is malformed {nfs_store_str}"
                )
                exit(1)
            if len(nfs_store_parts) == 1:
                directory = "/nix/store"
            else:
                directory = nfs_store_parts[1]
            kernel_params = f"nfs_store={ip}:{directory}"
        else:
            nfs = socket.getfqdn("nfs")
            store_path = get_store_path()
            kernel_params = f"nfs_store={nfs}:/export{store_path}"
        self.ctx.vlog(f" kernel_params: {kernel_params}")

        generate_kexec_scripts(self.ctx, flavour_kernel_params=kernel_params)


class G5kRamdiskFlavour(G5kKexecBasedFlavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.name = "g5k-ramdisk"


class G5kImageFlavour(G5kFlavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.name = "g5k-image"
        self.ask_before_kadeploy = True

    def launch(self, machine_file=None, kaenv_path=None, deploy_image_path=None):
        generate_kadeploy_envfile(
            self.ctx, kaenv_path=kaenv_path, deploy_image_path=deploy_image_path
        )

        remote_store_url = None
        if self.ctx.image_store_ssh:
            remote_store_url = f"ssh://{self.ctx.image_store_ssh}"
        image_path, use_image_store_ssh = realpath_from_store_remote(
            self.ctx, self.ctx.deployment_info["all"]["image"], remote_store_url
        )

        if deploy_image_path is None:
            user = os.environ["USER"]
            deploy_image_path = f"~{user}/public/nixos.tar.xz"

        if use_image_store_ssh:
            cmd_copy = "scp"
            image_path = f"{self.ctx.image_store_ssh}:{image_path}"
        else:
            cmd_copy = "cp"

        cmd_copy_image = f"{cmd_copy} {image_path} {deploy_image_path} && chmod 644 {deploy_image_path}"
        if (
            machine_file
            or self.ctx.machine_file
            or click.confirm(f"Do you want to copy image to {deploy_image_path} ?")
        ):
            try:
                subprocess.call(cmd_copy_image, shell=True)
            except Exception as ex:
                raise click.ClickException(f"Failed to copy image: {ex}")
        else:
            print(f"You can copy image with: {cmd_copy_image}")
        base_path = op.join(
            self.ctx.envdir,
            f"artifact/{self.ctx.composition_name}/{self.ctx.flavour.name}",
        )
        if kaenv_path is None:
            kaenv_path = op.join(base_path, "nixos.yaml")

        fqdn = socket.getfqdn()
        g5k_site = fqdn.split(".")[1]
        g5k_frontend = "f" + g5k_site

        ssh2frontend = ""
        if g5k_frontend != socket.gethostname():
            ssh2frontend = f"ssh {g5k_frontend} -t"

        cmd_kadeploy = (
            f"{ssh2frontend} kadeploy3 -a {kaenv_path} -f {self.ctx.machine_file}"
        )

        # g5k_frontend == socket.gethostname() and "OAR_NODEFILE" in os.environb:
        # cmd_kadeploy = f"kadeploy3 -a {kaenv_path} -f $OAR_NODEFILE"

        if not self.ask_before_kadeploy or click.confirm(
            f"Do you want to launch kadeploy: \n {cmd_kadeploy}"
        ):
            try:
                self.ctx.vlog(cmd_kadeploy)
                subprocess.call(cmd_kadeploy, shell=True)
            except Exception as ex:
                raise click.ClickException(f"Failed to execute kadeploy command: {ex}")
        else:
            print(f"You can kadeploy image with: {cmd_kadeploy}")

    def start(self, machine):
        if not self.ctx.no_start:
            print("Not Yet Implemented")
            exit(1)
        else:
            machine.start_process_shell(
                [
                    "ssh",
                    "-t",
                    "-o",
                    "StrictHostKeyChecking=no",
                    "-l",
                    "root",
                    machine.ip,
                ]
            )

    def ext_connect(self, user, node, execute=True, ssh_key_file=None):
        return ssh_connect(self.ctx, user, node, execute, ssh_key_file)



from pathlib import Path
import argparse
import ptpython.repl
import os
import time

from .logger import rootlog
from .driver import Driver


def generate_driver_symbols() -> None:
    """
    This generates a file with symbols of the test-driver code that can be used
    in user's test scripts. That list is then used by pyflakes to lint those
    scripts.
    """
    d = Driver([], [], "")
    test_symbols = d.test_symbols()
    with open("driver-symbols", "w") as fp:
        fp.write(",".join(test_symbols.keys()))



import click
import os
import os.path as op
import glob

from ..context import pass_context
from ..actions import read_deployment_info
from ..flavours import get_flavour_by_name


@click.command("stop")
@click.option(
    "-f",
    "--flavour",
    type=click.STRING,
    help="specify flavour",
)
@click.option(
    "-d",
    "--deployment-file",
    type=click.STRING,
    help="specify deployment",
)
@pass_context
def cli(ctx, flavour, deployment_file):
    """Stop Nixos composition."""
    ctx.log("Stopping")

    flavour_name = flavour

    deploy_path = op.join(ctx.envdir, "deploy")

    if not deployment_file:
        if flavour_name:
            search_path = f"{deploy_path}/*::{flavour_name}.json"
        else:
            search_path = f"{deploy_path}/*"

        deploy_paths = glob.glob(search_path)
        if not deploy_paths:
            raise click.ClickException("Failed to find last deployment file")

        last_deploy_path = max(
            deploy_paths,
            key=lambda x: os.stat(x, follow_symlinks=False).st_ctime,
        )

        deployment_file = last_deploy_path
        ctx.log("Use last deployment file:")
        ctx.glog(deployment_file)

        # ctx.composition_flavour_prefix = op.basename(last_deploy_path)

    splitted_filename = deployment_file.split("::")

    if len(splitted_filename) != 2:
        raise click.ClickException("Sorry, filename is malformed")

    splitted_last = splitted_filename[-1].split(".")
    if splitted_last[-1] != "json":
        raise click.ClickException("Sorry, filename is malformed")

    if flavour_name:
        assert flavour_name == splitted_last[0]
    else:
        flavour_name = splitted_last[0]

    ctx.flavour = get_flavour_by_name(flavour_name)(ctx)

    read_deployment_info(ctx, deployment_file)

    ctx.flavour.cleanup()



import os
import os.path as op
import shutil
import filecmp

import click


def touch(fname, times=None):
    dirname = "/".join(fname.split("/")[:-1])
    if not op.exists(dirname):
        os.makedirs(dirname)
    with open(fname, "a"):
        os.utime(fname, times)


def copy_file(srcname, dstname, preserve_symlinks=True):
    if preserve_symlinks and op.islink(srcname):
        if op.islink(dstname):
            os.unlink(dstname)
        else:
            os.remove(dstname)
        linkto = os.readlink(srcname)
        os.symlink(linkto, dstname)
    else:
        if op.islink(dstname):
            os.unlink(dstname)
        shutil.copy2(srcname, dstname)


def copy_tree(src, dest, overwrite=False, ignore_if_exists=[]):
    """
    Copy all files in the source path to the destination path.
    """
    if op.exists(dest) and not overwrite:
        raise click.ClickException("File exists : '%s'" % dest)
    create = click.style("   create", fg="green")
    # chmod = click.style("    chmod", fg="cyan")
    overwrite = click.style("overwrite", fg="yellow")
    identical = click.style("identical", fg="blue")
    ignore = click.style("   ignore", fg="magenta")
    cwd = os.getcwd() + "/"
    for path, dirs, files in os.walk(src):
        relative_path = path[len(src) :].lstrip(os.sep)
        if not op.exists(op.join(dest, relative_path)):
            os.mkdir(op.join(dest, relative_path))
        for filename in files:
            src_file_path = op.join(path, filename)
            dest_file_path = op.join(dest, relative_path, filename)
            if dest_file_path.startswith(cwd):
                fancy_relative_path = dest_file_path.replace(cwd, "")
            else:
                fancy_relative_path = dest_file_path
            if op.exists(dest_file_path):
                if filename in ignore_if_exists:
                    click.echo("   " + ignore + "  " + fancy_relative_path)
                elif filecmp.cmp(src_file_path, dest_file_path):
                    click.echo("   " + identical + "  " + fancy_relative_path)
                else:
                    click.echo("   " + overwrite + "  " + fancy_relative_path)
                    copy_file(src_file_path, dest_file_path)
            else:
                click.echo("   " + create + "  " + fancy_relative_path)
                copy_file(src_file_path, dest_file_path)



import click
import sys
import socket
from ..context import pass_context
from ..g5k import key_sleep_script, g5k_get_seed_store
from ..actions import install_nix_static, get_ip_ssh_port
from ..tools.nested_deployment import main as nested


def print_helper(ctx, options):
    option = options[0]
    if (option == "g5k_script") or (option == "g5k-script"):
        click.echo(key_sleep_script)
    elif option == "install-nix":
        install_nix_static(ctx)
    elif option == "ip":
        if len(options) > 1:
            ip, _ = get_ip_ssh_port(ctx, options[1])
            print(f"{ip}")
        else:
            ctx.elog("Host argument required")
            sys.exit(1)
    elif option == "ip_ssh_port":
        if len(options) > 1:
            ip, ssh_port = get_ip_ssh_port(ctx, options[1])
            print(f"{ip}:{ssh_port}")
        else:
            ctx.elog("Host argument required")
            sys.exit(1)
    elif option == "fqdn":
        if len(options) > 1:
            ip, _ = get_ip_ssh_port(ctx, options[1])
            fqdn = socket.getfqdn(ip)
            print(f"{fqdn}")
    elif option == "g5k-get-seed-store":
        if len(options) > 1:
            g5k_get_seed_store(ctx, options[1])
        else:
            g5k_get_seed_store(ctx)
    elif option == "nested":
        nested(options[1:])

    else:
        ctx.elog(f"Helper: {option} does not exist")
        sys.exit(1)


def print_helper_list(helper_options):
    print("g5k-script: print path to g5k_key_sleep_script Grid'5000 script")
    print("install-nix: install the nix command in ~/.local/bin")
    print(
        "g5k-get-seed-store: get an initial store default url: http://public.grenoble.grid5000.fr/~orichard/seed-nix-store.tgz"
    )
    print("ip <composition_hostname>: print hostname's ip address")
    print(
        "ip_ssh_port <composition_hostname>: print hostname's ip address and ssh port"
    )
    print("fqdn <composition_hostname>: print hostname's fully qualified domain name")
    print(
        "nested [options]: various file generation for nested deployement (-h for options details"
    )


@click.command("helper")
@click.option(
    "-l",
    "--list",
    # "--list-helpers",
    is_flag=True,
    help="List of available helpers",
)
@pass_context
@click.argument("options", nargs=-1)
def cli(ctx, options, list):
    """Specific and contextual helper information (e.g. g5k_script path for Grid'5000)
    Warning: Experimental command, may be removed in the future or change without backward compatibility care.
    """

    if options:
        print_helper(ctx, options)

    if list:
        print_helper_list(ctx)



import http.server
import sys
import os
import socket
import socketserver
import threading


class HTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=HTTPDaemon.directory, **kwargs)

    def log_message(self, format, *args):
        pass

    def log_error(self, format, *args):
        message = "%s - - [%s] %s\n" % (
            self.address_string(),
            self.log_date_time_string(),
            format % args,
        )
        if HTTPDaemon.ctx:
            HTTPDaemon.ctx.elog(message)
        else:
            sys.stderr.write(message)

    def do_GET(self):
        log_message = f"{self.client_address[0]}: HTTP GET {self.path}"
        if HTTPDaemon.ctx:
            HTTPDaemon.ctx.vlog(log_message)
        else:
            print(log_message)
        http.server.SimpleHTTPRequestHandler.do_GET(self)
        with HTTPDaemon.lock:
            HTTPDaemon.machines.append(self.client_address[0])
            # if len(HTTPDaemon.machines) == HTTPDaemon.expected_nb_machines:
            #    HTTPDaemon.get_done_event.set()


class HTTPDaemon:
    lock = threading.Lock()
    machines = []
    expected_nb_machines = 0
    # get_done_event = threading.Event()
    directory = ""
    ctx = None

    def __init__(self, ctx=None, port=0):
        HTTPDaemon.ctx = ctx
        self.httpd = socketserver.ThreadingTCPServer(("", port), HTTPRequestHandler)
        self.port = self.httpd.server_address[1]

        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # doesn't have to be reachable
        s.connect(("10.255.255.255", 1))
        self.ip = s.getsockname()[0]

        self.httpd_thread = threading.Thread(
            target=self.httpd.serve_forever, daemon=True
        )

    def start(self, expected_nb_machines=0, directory=os.getcwd()):
        HTTPDaemon.expected_nb_machines = expected_nb_machines
        HTTPDaemon.directory = directory

        self.httpd_thread.start()

        # HTTPDaemon.get_done_event.wait()
        # self.httpd.shutdown()

    def stop(self):
        self.httpd.shutdown()

    # def wait(self, nb_machines):
    #    pass



import json
import os
import os.path as op
import glob
import socket
import sys
import shutil
import subprocess
import time
import base64
import click
import signal
import psutil
import itertools
import ipaddress
import urllib.request

from .tools.kataract import generate_scp_tasks, exec_kataract_tasks

# from .default_role import DefaultRole #TODO


##
# Helper function to determine fstype
#
def get_fs_type(path):
    root_type = ""
    for part in psutil.disk_partitions(True):
        if part.mountpoint == "/":
            root_type = part.fstype
            continue
        if path.startswith(part.mountpoint):
            return part.fstype
    return root_type


##
# Determine nix store location
#
def nix_store_location(ctx):
    for store_path in ["/nix"] + ctx.alternative_stores:
        if op.exists(store_path):
            return f"{store_path}/store"
    ctx.wlog("Failed to find nix store location")
    return ""


##
# Retrieve from path from different store location if needed
#
def realpath_from_store_core(ctx, path, include_prefix_store=False):
    p = op.realpath(path)
    potential_store_paths = ctx.alternative_stores + [
        op.join(ctx.envdir, "artifact/nix")
    ]
    for store_path in potential_store_paths:
        new_p = f"{store_path}{p[4:]}"
        if op.exists(new_p):
            if include_prefix_store:
                return new_p, store_path
            else:
                return new_p, None
    if op.exists(p):
        return p, None
    return None, None


def realpath_prefix_from_store(ctx, path):
    realpath, prefix_store = realpath_from_store_core(ctx, path, True)
    if not realpath:
        ctx.elog(f"{path} does not exist in standard store or alternate")
        sys.exit(1)
    return realpath, prefix_store


def realpath_from_store(ctx, path):
    realpath, _ = realpath_from_store_core(ctx, path)
    if not realpath:
        ctx.elog(f"{path} does not exist in standard store or alternate")
        sys.exit(1)
    return realpath


def realpath_from_store_remote(ctx, path, remote_store_url=None):
    realpath, _ = realpath_from_store_core(ctx, path)
    if realpath is not None:
        return realpath, False
    if remote_store_url:
        cmd = ["ssh"]
        if remote_store_url[:6] == "ssh://":
            s = remote_store_url[6:].split(":")
            if len(s) == 2:
                cmd += ["-p", s[1]]
            cmd += [s[0], "ls", path]
            ctx.vlog(f"Check path in remote store: {cmd}")
            retcode = subprocess.run(cmd).returncode
            if retcode:
                ctx.vlog(f"Remote store check of path failed, return code: {retcode}")
            else:
                return path, True
        else:
            ctx.elog(
                f"Remote store url is not supported or malformed, only ssh://username@host:port is support: {remote_store_url}"
            )
            sys.exit(1)

    ctx.elog(f"{path} does not exist in standard, alternate or remote stores")
    sys.exit(1)


##
# Generate/manipulate/copy deploy, compose files
#
def get_deployment_file(ctx, deployment_file, flavour, tag):
    def exit_is_not_file(f):
        if not op.isfile(f):
            ctx.elog(f"{f} is not a file, deployment_file option must be provided")
            sys.exit(1)

    deploy_dir = op.join(ctx.envdir, "deploy")
    if not deployment_file:
        if not op.isdir(deploy_dir):
            ctx.elog("Failed to find deploy directory, is composition started ?")
            sys.exit(1)
        deployment_file = max(glob.glob(f"{deploy_dir}/*::*"), key=op.getctime)
        exit_is_not_file(deployment_file)
        return deployment_file
    else:
        if tag:
            if flavour:
                search_path = f"{deploy_dir}/*::{flavour}::{tag}.json"
            else:
                search_path = f"{deploy_dir}/*::{tag}.json"

            deployment_paths = glob.glob(search_path)
            if not deployment_paths:
                raise click.ClickException("Failed to find last deployment")

            deployment_file = max(
                deployment_paths,
                key=lambda x: os.stat(x, follow_symlinks=False).st_ctime,
            )
            return deployment_file
        else:
            base_deployment_file = deployment_file
            if op.exists(deployment_file):
                exit_is_not_file(deployment_file)
                return deployment_file
            else:
                deployment_file = op.join(
                    op.join(ctx.envdir, "deploy"), deployment_file
                )
                if op.exists(deployment_file):
                    exit_is_not_file(deployment_file)
                    return deployment_file
                else:
                    ctx.elog(f"{base_deployment_file} not found")
                    sys.exit(1)


def read_deployment_info(ctx, deployment_file=None, flavour=None, tag=None):
    ctx.deployment_filename = get_deployment_file(ctx, deployment_file, flavour, tag)
    with open(ctx.deployment_filename, "r") as f:
        deployment_info = json.load(f)
    ctx.deployment_info = deployment_info
    if "composition" in deployment_info:
        composition_name = deployment_info["composition"]
        if not ctx.composition_name:
            ctx.composition_name = composition_name
        elif ctx.composition_name != composition_name:
            ctx.wlog(
                "Composition from built ({ctx.composition_name}) is different from deployment ({composition_name})"
            )
    return


def read_deployment_info_str(ctx, deployment_file=None):
    ctx.deployment_filename = get_deployment_file(ctx, deployment_file)
    with open(ctx.deployment_filename, "r") as f:
        deployment_info_str = f.read()
    return deployment_info_str


def read_test_script(ctx, compose_info_or_str):
    if not compose_info_or_str:
        return None
    if isinstance(compose_info_or_str, str):
        filename = compose_info_or_str
    elif "test_script" in compose_info_or_str:
        filename = compose_info_or_str["test_script"]
    else:
        return None
    with open(realpath_from_store(ctx, filename), "r") as f:
        test_script = f.read()
        return test_script


def read_compose_info(ctx):
    if not op.isfile(ctx.compose_info_file):
        raise click.ClickException(f"{ctx.compose_info_file} does not exist")
    with open(ctx.compose_info_file, "r") as f:
        compose_info = json.load(f)

    if "compositions_info" in compose_info:
        ctx.compositions_info = compose_info

        if len(compose_info["compositions_info"]) > 1:
            ctx.multiple_compositions = True
            ctx.vlog(
                f"Image with multiple compositions is detected, selected composition: {ctx.composition_name}"
            )

        if (
            ctx.composition_name
            and ctx.composition_name not in ctx.compositions_info["compositions_info"]
        ):
            ctx.elog(
                f'Composition named: "{ctx.composition_name}" is not in {ctx.compose_info_file}'
            )
            if ctx.composition_name == ctx.composition_basename_file:
                ctx.elog(
                    "When image contains multiple compositions and without default, the one to launch must by indicated via composition name option with name as prefix followed by file's label separated by ::  (e.g. -c foo::g5k-ramdisk). The default if present, is composition with its name equals to file's label."
                )
            sys.exit(1)
        if ctx.composition_name:
            compose_info = ctx.compositions_info["compositions_info"][
                ctx.composition_name
            ]

        if "all" not in compose_info:
            compose_info["all"] = ctx.compositions_info["all"]
            compose_info["flavour"] = ctx.compositions_info["flavour"]
            if "compositions_info_path" in ctx.compositions_info:
                compose_info["compositions_info_path"] = ctx.compositions_info[
                    "compositions_info_path"
                ]

        compose_flavour_name = ctx.compositions_info["flavour"]["name"]

        if ctx.flavour and ctx.flavour.name != compose_flavour_name:
            raise click.ClickException(
                f"Selected flavour ({ctx.flavour.name}) differs from compose info ({compose_flavour_name})"
            )

    ctx.compose_info = compose_info
    return


def get_machine_from_file(ctx):
    machines = [machine.rstrip() for machine in open(ctx.machine_file, "r")]
    if not machines:
        ctx.elog(f"Machine file '{ctx.machine_file}' is empty")
        sys.exit(1)
    ctx.machine_names_from_file = machines  # CHECK IT USED ?
    translate_hosts2ip(ctx, machines)
    print(ctx.ip_addresses, ctx.host2ip_address)


def translate_hosts2ip(ctx, hosts):
    for host in hosts:
        if host and (host not in ctx.host2ip_address):
            ip = socket.gethostbyname_ex(host)[2][0]
            ctx.host2ip_address[host] = ip
            ctx.ip_addresses.append(ip)
    return


def populate_deployment_vm_by_ip(ctx, roles_info, roles_distribution):
    roles_distribution = health_check_roles_distribution(
        ctx, roles_info, roles_distribution
    )
    i = 1
    deployment = {}
    ips = []
    for role, v in roles_info.items():
        for hostname in roles_distribution[role]:
            ip = "192.168.1.{}".format(i)
            ips.append(ip)
            # deployment[ip] = {"role": role, "vm_id": i}
            deployment[ip] = {
                "role": role,
                "vm_id": i,
                "host": hostname,
            }
            if isinstance(v, str):
                deployment[ip]["toplevel"] = v
            else:
                # backward compatibility, deprecated
                deployment[ip]["init"] = v["init"]
            i = i + 1

    return deployment, ips


def health_check_roles_distribution(ctx, roles_info, roles_distribution_in, ips=None):
    roles_distribution = {}
    # if isinstance(roles_info, list):
    #     roles = roles_info
    # else:
    #     roles = roles_info.keys()
    for role in roles_info.keys():
        if role in roles_distribution_in:
            roles_distribution[role] = roles_distribution_in[role]
        elif (
            "roles_distribution" in ctx.compose_info
            and role in ctx.compose_info["roles_distribution"]
        ):
            hosts = ctx.compose_info["roles_distribution"][role]
            if isinstance(hosts, int):
                hosts = [f"{role}{i}" for i in range(1, hosts + 1)]
            # TODO REMOVE after test
            # if isinstance(hosts, list):
            #     try:
            #         quantity = int(hosts)
            #         hosts = [f"{role}{i}" for i in range(1, quantity + 1)]
            #     except ValueError:
            #         pass
            roles_distribution[role] = hosts
        else:
            roles_distribution[role] = [role]

    # TODO
    # - if ips is present and ips number lower than host number
    # - if ips is present and ips number greater than host number
    #      do we manage default role to padding "--padding-role options ?"
    # manage defaultRole / mini in yaml/ default distribution role in nix composition

    # if len(roles_quantities_in) == 0:
    #     # If no info we take one node per role
    #     roles_quantities = {role: [role] for role in nodes_info.keys()}
    #     if ips:
    #         nb_nodes = len(ips) - len(nodes_info.keys())
    #         if nb_nodes >= 0 and "node" in roles_quantities:
    #             ctx.vlog("Apply node role as default on ")
    #             roles_quantities["node"] = [
    #                 f"node{i}" for i in range(1, (nb_nodes + 2))
    #             ]
    # else:
    #     sum_nb_asked_machines = 0
    #     for role in roles_quantities_in:
    #         if type(roles_quantities_in[role]) == int:
    #             sum_nb_asked_machines += roles_quantities_in[role]
    #         elif type(roles_quantities_in[role]) == list:
    #             sum_nb_asked_machines += len(roles_quantities_in[role])
    #         else:
    #             pass
    #     if ips is not None:
    #         remaining_available_machines = len(ips) - sum_nb_asked_machines
    #     else:
    #         remaining_available_machines = -1

    #     # Step 1: if the user only gave the number of nodes of the roles
    #     for role in roles_quantities_in:
    #         if type(roles_quantities_in[role]) == int:
    #             nb_nodes = roles_quantities_in[role]
    #             if nb_nodes == 1:
    #                 roles_quantities[role] = [f"{role}"]
    #             else:
    #                 roles_quantities[role] = [
    #                     f"{role}{i}" for i in range(1, nb_nodes + 1)
    #                 ]
    #         elif type(roles_quantities_in[role]) == DefaultRole:
    #             nb_min_nodes = roles_quantities_in[role].nb_min_nodes
    #             if ips is None:
    #                 # in the case of VMs we are not limited
    #                 # so we take the min nb of nodes
    #                 remaining_available_machines = nb_min_nodes
    #             if remaining_available_machines < nb_min_nodes:
    #                 raise Exception(
    #                     f"Not enough nodes to satisfy default role {role} ({remaining_available_machines} available for {nb_min_nodes} asked)"
    #                 )
    #             if remaining_available_machines == 1:
    #                 roles_quantities[role] = [f"{role}"]
    #             else:
    #                 roles_quantities[role] = [
    #                     f"{role}{i}" for i in range(1, remaining_available_machines + 1)
    #                 ]
    #         else:
    #             if not ips:
    #                 raise Exception(
    #                     "Number of ip_address must be known with 'remaining' as role's number"
    #                 )
    #             if remaining_role:
    #                 raise Exception(
    #                     f"Role for remaining nodes is already set: {remaining_role}/{role}"
    #                 )
    #             else:
    #                 remaining_role = role

    # Step 2: add remainings and check that we do not have any conflict on the hostnames
    all_hostnames = list(itertools.chain.from_iterable(roles_distribution.values()))
    set_hostnames = set(all_hostnames)
    if len(all_hostnames) != len(set_hostnames):
        raise Exception("Conflict in the naming of the nodes")

    return roles_distribution


def populate_deployment_ips(ctx, roles_info, ips, roles_distribution):
    roles_distribution = health_check_roles_distribution(
        ctx, roles_info, roles_distribution, ips
    )
    i = 0
    deployment = {}
    for role, v in roles_info.items():
        if role not in roles_distribution:
            ctx.elog(f"role: {role} not found in roles-distribution file")
            exit(1)
        for hostname in roles_distribution[role]:
            try:
                ip = ips[i]
            except IndexError as e:
                ctx.elog(f"Not enough nodes are available for the deployment: {e}")
                exit(1)
            # TODO Ugly need core refactoring to remove it
            if hasattr(ctx.flavour, "host_info"):
                deployment[ip] = ctx.flavour.host_info(role, hostname, v)
            else:
                deployment[ip] = {"role": role, "host": hostname}
                if isinstance(v, str):
                    deployment[ip]["toplevel"] = v
                else:
                    # backward compatibility, deprecated
                    deployment[ip]["init"] = v["init"]
            i = i + 1
    return deployment


def generate_deployment_info(ctx, ssh_pub_key_file=None):
    if not ctx.compose_info:
        read_compose_info(ctx)

    if not ssh_pub_key_file:
        ssh_pub_key_file = os.environ["HOME"] + "/.ssh/id_rsa.pub"
    with open(ssh_pub_key_file, "r") as f:
        sshkey_pub = f.read().rstrip()

    # if ctx.multiple_compositions:  :: TO REMOVE ???
    #    roles = ctx.compose_info["roles"]
    if ctx.ip_addresses:
        deployment = populate_deployment_ips(
            ctx, ctx.compose_info["roles"], ctx.ip_addresses, ctx.roles_distribution
        )
    else:
        deployment_base, ctx.ip_addresses = populate_deployment_vm_by_ip(
            ctx, ctx.compose_info["roles"], ctx.roles_distribution
        )
        deployment = {}
        for k, v in deployment_base.items():
            s = {"role": v["role"], "host": v["host"], "vm_id": v["vm_id"]}
            if "toplevel" in v:
                s["toplevel"] = v["toplevel"]
            else:
                # backward compatibility, deprecated
                s["init"] = v["init"] if "host" in v else v["role"]
            deployment[k] = s

    deployment = {
        "ssh_key.pub": sshkey_pub,
        "deployment": deployment,
    }

    if "all" in ctx.compose_info:
        deployment["all"] = ctx.compose_info["all"]

    if "compositions_info_path" in ctx.compose_info:
        deployment["compositions_info_path"] = ctx.compose_info[
            "compositions_info_path"
        ]

    if ctx.composition_name:
        deployment["composition"] = ctx.composition_name

    # Add user, used to determine nfs mount path on Grid'5000 by example
    # TODO: add option to override this (
    deployment["user"] = os.environ["USER"]

    # for k in ["all", "flavour"]:
    #    if k in compose_info:
    #        deployment[k] = compose_info[k]

    # If there is too much nodes httpd must used to tranfert deployment info,
    # due to kernel parameter size limit (deployment_info_b64 certainly will exceed it)
    if len(deployment["deployment"]) > 4:
        ctx.use_httpd = True

    json_deployment = json.dumps(deployment, indent=2)

    deploy_dir = op.join(ctx.envdir, "deploy")
    if not op.exists(deploy_dir):
        create = click.style("   create", fg="green")
        ctx.log("   " + create + "  " + deploy_dir)
        os.mkdir(deploy_dir)

    ctx.deployment_filename = op.join(
        deploy_dir, f"{ctx.composition_flavour_prefix}.json"
    )
    with open(ctx.deployment_filename, "w") as outfile:
        outfile.write(json_deployment)

    if "parameters" in ctx.deployment_info:
        deployment["parameters"] = ctx.deployment_info["parameters"]

    ctx.deployment_info = deployment

    return


def generate_kexec_scripts(ctx, flavour_kernel_params=""):
    if ctx.use_httpd:
        base_url = f"http://{ctx.httpd.ip}:{ctx.httpd.port}"
        deploy_info_src = f"{base_url}/deploy/{ctx.composition_flavour_prefix}.json"
    else:
        generate_deploy_info_b64(ctx)
        deploy_info_src = ctx.deployment_info_b64

    base_path = op.join(
        ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
    )
    kexec_scripts_path = op.join(base_path, "kexec_scripts")
    os.makedirs(kexec_scripts_path, mode=0o700, exist_ok=True)

    kernel_params = ""
    if ctx.kernel_params:
        kernel_params = ctx.kernel_params

    if "all" in ctx.deployment_info:
        kernel_path = realpath_from_store(ctx, ctx.deployment_info["all"]["kernel"])
        initrd_path = realpath_from_store(ctx, ctx.deployment_info["all"]["initrd"])

        kexec_args = "-l $KERNEL --initrd=$INITRD "
        kexec_args += rf'--append="deploy={deploy_info_src} console=tty0 console=ttyS0,115200 check_kernel_initrd {flavour_kernel_params} {kernel_params}"'
        script_path = op.join(kexec_scripts_path, "kexec.sh")
        with open(script_path, "w") as kexec_script:
            kexec_script.write("#!/usr/bin/env bash\n")
            kexec_script.write(": ${KERNEL:=" + kernel_path + "}\n")
            kexec_script.write(": ${INITRD:=" + initrd_path + "}\n")
            kexec_script.write(f"kexec {kexec_args}\n")
            kexec_script.write("kexec -e\n")
        os.chmod(script_path, 0o755)
    else:
        for ip, v in ctx.deployment_info["deployment"].items():
            role = v["role"]
            kernel_path = f"{base_path}/kernel_{role}"
            initrd_path = f"{base_path}/initrd_{role}"
            if "toplevel" in v:
                init_path = f"{v['tolelevel']}/init"
            else:
                # backward compatibility, deprecated
                init_path = v["init"]
            kexec_args = f"-l {kernel_path} --initrd={initrd_path} "
            kexec_args += rf'--append="init={init_path} deploy={deploy_info_src} console=tty0 console=ttyS0,115200 check_kernel_initrd {flavour_kernel_params} {kernel_params}"'
            script_path = op.join(kexec_scripts_path, f"kexec_{role}.sh")
            with open(script_path, "w") as kexec_script:
                kexec_script.write("#!/usr/bin/env bash\n")
                kexec_script.write(f"kexec {kexec_args}\n")
                kexec_script.write("kexec -e\n")

            os.chmod(script_path, 0o755)


def generate_deploy_info_b64(ctx):
    deployment_info = {
        k: ctx.deployment_info[k]
        for k in [n for n in ctx.deployment_info.keys() if n != "deployment"]
    }

    deployment = {
        k: {"role": v["role"], "host": v["host"] if "host" in v else v["role"]}
        for k, v in ctx.deployment_info["deployment"].items()
    }

    # TODO: function to add multiple hostnames with same role
    # deployment = { k: {"role": v["role"], "host": "yopXXX"} for k,v in ctx.deployment_info["deployment"].items()}

    deployment_info["deployment"] = deployment

    ctx.vlog(f"deploy info \n{deployment_info}")
    deployment_info_str = json.dumps(deployment_info)

    ctx.deployment_info_b64 = base64.b64encode(deployment_info_str.encode()).decode()

    if len(ctx.deployment_info_b64) > (4096 - 256):
        ctx.log(
            "The base64 encoded deploy data is too large: use an http server to serve it"
        )
        sys.exit(1)
    return


def artifact_copy_all_kernel_initrd(ctx):
    read_compose_info(ctx)
    compose_info_all = ctx.compose_info["all"]
    kernel_path = compose_info_all["kernel"]
    initrd_path = compose_info_all["initrd"]

    for store_path in [kernel_path, initrd_path]:
        artifact_path = op.join(ctx.envdir, f"artifact{store_path}")
        if not op.exists(artifact_path):
            artifact_dir = op.dirname(artifact_path)
            if not op.exists(artifact_dir):
                os.makedirs(artifact_dir, mode=0o700, exist_ok=True)
            shutil.copy(store_path, op.join(ctx.envdir, f"artifact{store_path}"))


# def copy_result_from_store(ctx):

#     if not ctx.compose_info:
#         read_compose_info(ctx)

#     if ctx.multiple_compositions:
#         composition_directory = "::"
#     else:
#         composition_directory = ctx.composition_name

#     artifact_path = op.join(
#         ctx.envdir, f"artifact/{composition_directory}/{ctx.flavour_name}"
#     )
#     if not op.exists(artifact_path):
#         create = click.style("   create", fg="green")
#         ctx.log("   " + create + "  " + artifact_path)
#         pathlib.Path(artifact_path).mkdir(parents=True, exist_ok=True)

#     if ctx.multiple_compositions:
#         compose_info = ctx.compositions_info
#     else:
#         compose_info = ctx.compose_info

#     new_compose_info = compose_info.copy()

#     if "all" in compose_info:
#         for target in ["kernel", "initrd", "qemu_script", "image"]:
#             if target in compose_info["all"]:
#                 new_target = op.join(artifact_path, target)
#                 shutil.copy(compose_info["all"][target], new_target)
#                 os.chmod(new_target, 0o644)
#                 new_compose_info["all"][target] = new_target
#     else:
#         for r, v in compose_info["nodes"].items():
#             for target in ["kernel", "initrd", "qemu_script", "image"]:
#                 if target in v:
#                     new_target = op.join(artifact_path, f"{target}_{r}")
#                     shutil.copy(v[target], new_target)
#                     os.chmod(new_target, 0o644)
#                     new_compose_info["nodes"][target] = new_target

#     if "test_script" in compose_info:
#         new_target = op.join(artifact_path, "test_script")
#         shutil.copy(compose_info["test_script"], new_target)
#         os.chmod(new_target, 0o644)
#         new_compose_info["test_script"] = new_target

#     # save new updated compose_info
#     json_new_compose_info = json.dumps(new_compose_info, indent=2)
#     with open(
#         op.join(ctx.envdir, f"build/{ctx.composition_flavour_prefix}::artifact"), "w",
#     ) as outfile:
#         outfile.write(json_new_compose_info)

#     ctx.compose_info = new_compose_info
#     ctx.log("Copy from store: " + click.style("completed", fg="green"))


##
#  Operation: connect, launch, wait_ssh_port,
#


def launch_ssh_kexec(ctx, ip=None, debug=False, push_path=None):
    if ctx.show_spinner:
        ctx.spinner.start("Launching remote kexec(s)")
    else:
        ctx.log("Launching remote kexec(s)")

    if "all" in ctx.deployment_info:
        if push_path:
            kexec_script = f"{push_path}/kexec.sh"
            ki = f"KERNEL={push_path}kernel INITRD={push_path}initrd"
            user = "root@"
        else:
            base_path = op.join(
                ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
            )
            kexec_script = op.join(base_path, "kexec_scripts/kexec.sh")
            ki = ""
            user = ""
        # USELESS ? TOREMOVE ?
        # if ctx.sudo:
        #     sudo = f"SUDO={ctx.sudo}"
        # else:
        #     sudo = ""

        if "DEBUG_STAGE1" in os.environ or debug:
            # debug_stage1 = os.environ["DEBUG_STAGE1"]
            # TODO
            ctx.wlog(
                "Machine selection not yet supported within ssh_kexec. Debug_stage1 will apply to all."
            )
            ki = f" {ki} DEBUG_INITRD=boot.debug1mounts "

        def one_ssh_kexec(ip_addr):
            ssh_cmd = f'{ctx.ssh} {user}{ip_addr} "screen -dm bash -c \\" {ki} {kexec_script}\\""'
            ctx.vlog(ssh_cmd)
            subprocess.call(ssh_cmd, shell=True)

        if ip:
            one_ssh_kexec(ip)
        else:
            for ip in ctx.deployment_info["deployment"].keys():
                one_ssh_kexec(ip)
    else:
        raise Exception("Sorry, only all-in-one image version is supported up to now")

    if ctx.show_spinner:
        ctx.spinner.succeed("Remote kexec(s) launched")


def wait_ssh_ports(ctx, ips=None):
    if not ctx.show_spinner:
        ctx.log("Waiting ssh ports:")
    if not ips:
        ips = ctx.ip_addresses
    nb_ips = len(ips)

    nb_ssh_port = 0
    waiting_ssh_ports_cmd = (
        f"nmap -p22 -Pn {' '.join(ips)} -oG - | grep '22/open' | wc -l"
    )
    ctx.vlog(waiting_ssh_ports_cmd)

    if ctx.show_spinner:
        ctx.spinner.start(f"Waiting ssh ports, opened: 0/{nb_ips}")

    while nb_ssh_port != nb_ips:
        output = subprocess.check_output(waiting_ssh_ports_cmd, shell=True)
        nb_ssh_port = int(output.rstrip().decode())
        if ctx.show_spinner:
            ctx.spinner.text(
                "Opened ssh ports: {}/{} ({:.1f}s)".format(
                    nb_ssh_port, nb_ips, ctx.elapsed_time()
                )
            )
        time.sleep(0.25)
    if ctx.show_spinner:
        ctx.spinner.succeed("Deployment taken {:.1f} sec".format(ctx.elapsed_time()))
    else:
        ctx.vlog("Deployment took {:.1f}s".format(ctx.elapsed_time()))


def push_on_machines(ctx, push_path=None):
    if "all" not in ctx.deployment_info:
        raise Exception("Sorry, only all-in-one image version is supported up to now")

    kernel = realpath_from_store(ctx, ctx.deployment_info["all"]["kernel"])
    initrd = realpath_from_store(ctx, ctx.deployment_info["all"]["initrd"])

    base_path = ctx.envdir

    # if ctx.multiple_compositions:
    #    subpath = "::"
    # else:
    # subpath = ctx.composition_name
    # base_path = op.join(
    #     ctx.envdir, f"artifact/{subpath}/{ctx.flavour_name}"
    # )
    subpath = ctx.composition_name
    base_path = op.join(ctx.envdir, f"artifact/{subpath}/{ctx.flavour_name}")

    kexec_script = op.join(base_path, "kexec_scripts/kexec.sh")

    ctx.vlog(
        f"push kernel, initrd, kexec_script on {ctx.ip_addresses} with scp executed concurrently"
    )
    for file_input in [kernel, initrd, kexec_script]:
        ctx.vlog(f"push: {file_input}")
        tasks_cmd = generate_scp_tasks(
            ctx.ip_addresses, file_input, push_path, scp="scp", user="root"
        )
        exec_kataract_tasks(tasks_cmd, elog=ctx.elog, vlog=ctx.vlog)

    # if shutil.which("kastafior"):
    #    raise NotImplementedError
    # elif shutil.which("kaput"):
    #     ctx.vlog("push kernel, initrd, kexec_script on hosts with kaput")
    #     joined_ip_addresses = ",".join(ctx.ip_addresses)
    #     for f in [kernel, initrd, kexec_script]:
    #         kaput_cmd = f"kaput -l root -n {joined_ip_addresses} {f} {push_path}"
    #         ctx.vlog(kaput_cmd)
    #         subprocess.call(kaput_cmd, shell=True)
    # else:
    #     for ip_address in ctx.ip_addresses:
    #         ctx.vlog(f"push kernel, initrd, kexec_script to {ip_address}")
    #         for f in [kernel, initrd, kexec_script]:
    #             subprocess.call(
    #                 f"scp {f} root@{ip_address}:{push_path}", shell=True
    #             )


def get_ip_ssh_port(ctx, host):
    """Retrieve host's ip address and ssh port from deployment info"""
    ip_out = ""
    ssh_port = 22
    if not ctx.deployment_info:
        read_deployment_info(ctx)
    for ip, v in ctx.deployment_info["deployment"].items():
        if v["host"] == host:
            ip_out = ip
            if "vm_id" in v:
                ip_out = "127.0.0.1"
                ssh_port = 22021 + int(v["vm_id"])
            break
    return (ip_out, ssh_port)


def ssh_connect(ctx, user, host, execute=True, ssh_key_file=None):
    ip, ssh_port = get_ip_ssh_port(ctx, host)
    ssh_key_option = (
        ""
        if ssh_key_file is None
        else "-o IdentitiesOnly=yes -i " + os.path.realpath(ssh_key_file)
    )

    ssh_cmd = (
        f"ssh {ssh_key_option} -o StrictHostKeyChecking=no -o LogLevel=ERROR"
        f" -l {user} -p {ssh_port} {ip}"
    )

    if execute:
        return_code = subprocess.run(ssh_cmd, shell=True).returncode

        if return_code:
            ctx.wlog(f"SSH exit code is not null: {return_code}")
        sys.exit(return_code)
    else:
        return ssh_cmd


NB_PANES_2_GEOMETRY = ["1", "1+1", "1+2", "2+2", "2+3", "3+3", "3+4", "4+4"]


def connect_tmux(
    ctx, user, nodes, ssh_key_file, pane_console, geometry, window_name="nxc"
):
    if not nodes:
        deploy = ctx.deployment_info["deployment"]
        node = (list(deploy.keys()))[0]
        try:
            ipaddress.ip_address(node)
            nodes = [v["host"] for v in deploy.values()]
        except ValueError:
            nodes = list(deploy.keys())

    connect_cmds = [
        ctx.flavour.ext_connect(user, node, False, ssh_key_file) for node in nodes
    ]

    console = 0
    if pane_console:
        console = 1

    if not geometry:
        geometry = ""
        nb_panes = len(nodes)
        if pane_console:
            nb_panes += 1
        geometry = NB_PANES_2_GEOMETRY[min(7, nb_panes - 1)]

    # translate geometry
    if "+" in geometry and "*" in geometry:
        raise Exception("Mixing + and * in geometry is not supported")
    if "+" in geometry:
        splitw = [int(i) for i in geometry.split("+")]
        splitw.reverse()
    elif "*" in geometry:
        g = geometry.split("*")
        splitw = [int(g[1]) for i in range(int(g[0]))]
    else:
        splitw = [int(geometry)]

    nb_panes = sum(splitw)
    nb_splitv = len(splitw)

    ctx.vlog(f"geometry: {geometry}")
    ctx.vlog(f"splitw: {splitw}")
    ctx.vlog(f"nb_panes: {nb_panes}")

    # prepare commands
    base_cmds = ["bash" for i in range(nb_panes)]

    nb_connect_cmds = len(connect_cmds)
    if (nb_connect_cmds + console) > nb_panes:
        connect_cmds = connect_cmds[: nb_panes - console]
    cmds = base_cmds[: nb_panes - nb_connect_cmds] + connect_cmds

    ctx.vlog(f"cmds: {cmds}")

    cmds.reverse()

    if "TMUX" not in os.environ:
        cmd = "tmux new -d"
        subprocess.call(cmd, shell=True)

    i = 0

    cmd = f"tmux new-window -n {window_name} -d"
    if not pane_console:
        cmd += f" {cmds[i]}"
        i += 1
    subprocess.call(cmd, shell=True)

    # cmd = f'tmux splitw -h -p 50 -t {window_name}.0 "{ssh_cmds[0]}"'
    # subprocess.call(cmd, shell=True)

    for v in range(nb_splitv):
        ratio_h = round(100.0 / (nb_splitv - v))
        if ratio_h != 100:
            cmd = f"tmux splitw -h -p {ratio_h} -t {window_name}.0 {cmds[i]}"
            pane0 = 1
            # print("vertical |", cmd)
            # print(round(100 * (1.0 / (nb_splitv - v))))
            subprocess.call(cmd, shell=True)
            i += 1
        else:
            pane0 = 0

        for h in range(splitw[v] - 1):
            ratio_v = round(100 * (1 - (1.0 / (splitw[v] - h))))
            cmd = f"tmux splitw -v -p {ratio_v} -t {window_name}.{h+pane0} {cmds[i]}"
            # print(cmd)
            subprocess.call(cmd, shell=True)
            i += 1

    cmd = f"tmux select-pane -t {window_name}.0"
    subprocess.call(cmd, shell=True)

    cmd = f"tmux select-window -t {window_name}"
    subprocess.call(cmd, shell=True)

    if "TMUX" not in os.environ:
        cmd = "tmux attach"
        subprocess.call(cmd, shell=True)


# Helpers
def kill_proc_tree(
    pid, sig=signal.SIGTERM, include_parent=True, timeout=None, on_terminate=None
):
    """Kill a process tree (including grandchildren) with signal
    "sig" and return a (gone, still_alive) tuple.
    "on_terminate", if specified, is a callback function which is
    called as soon as a child terminates.
    From: https://psutil.readthedocs.io/en/latest/#kill-process-tree
    """
    assert pid != os.getpid(), "won't kill myself"
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    if include_parent:
        children.append(parent)
    for p in children:
        try:
            p.send_signal(sig)
        except psutil.NoSuchProcess:
            pass
    gone, alive = psutil.wait_procs(children, timeout=timeout, callback=on_terminate)
    return (gone, alive)


# test nix available
def get_nix_command(ctx):
    nix_cmd = ["nix"]

    if not shutil.which("nix"):
        local_bin_nix = "{os.environ['HOME']}/.local/bin/nix"
        if not op.exists(local_bin_nix):
            ctx.elog(
                "Nix not found, it can by installed in $HOME/.local/bin with command: nxc helper install-nix"
            )
            sys.exit(1)
        else:
            nix_cmd = [local_bin_nix]

    nix_cmd += ["--extra-experimental-features", "nix-command flakes"]
    return nix_cmd


# get Nix-static
def install_nix_static(
    ctx,
    archi="x86_64",
    local_bin_path=f"{os.environ['HOME']}/.local/bin",
):
    if ctx.show_spinner:
        ctx.spinner.start("Installing Nix")
    else:
        ctx.log("Installing Nix...")

    if not op.exists(local_bin_path):
        os.makedirs(local_bin_path)
    nix_path = op.join(local_bin_path, "nix")

    remote_nix_filename = ""
    with urllib.request.urlopen(
        f"https://gitlab.inria.fr/nixos-compose/nix-static/-/raw/main/bin/nix-{archi}-linux-static"
    ) as response:
        remote_nix_filename = response.read().decode("utf-8")

    ctx.log(f"Retrieving... {remote_nix_filename}")

    with urllib.request.urlopen(
        f"https://gitlab.inria.fr/nixos-compose/nix-static/-/raw/main/bin/{remote_nix_filename}"
    ) as response:
        with open(nix_path, "wb") as output_file:
            shutil.copyfileobj(response, output_file)

    os.chmod(nix_path, 0o755)

    # create store directory to satisfy autodetection in build phase
    os.makedirs(f"{os.environ['HOME']}/.local/share/nix/root/nix/store")

    if ctx.show_spinner:
        ctx.spinner.succeed("Nix installed")
    else:
        ctx.log("Done")



from colorama import Style
from contextlib import contextmanager
from typing import Any, Dict, Iterator
from queue import Queue, Empty
from xml.sax.saxutils import XMLGenerator
import codecs
import os
import sys
import time
import unicodedata


class Logger:
    def __init__(self) -> None:
        self.logfile = os.environ.get("LOGFILE", "/dev/null")
        self.logfile_handle = codecs.open(self.logfile, "wb")
        self.xml = XMLGenerator(self.logfile_handle, encoding="utf-8")
        self.queue: "Queue[Dict[str, str]]" = Queue()

        self.xml.startDocument()
        self.xml.startElement("logfile", attrs={})

        self._print_serial_logs = True

    @staticmethod
    def _eprint(*args: object, **kwargs: Any) -> None:
        print(*args, file=sys.stderr, **kwargs)

    def close(self) -> None:
        self.xml.endElement("logfile")
        self.xml.endDocument()
        self.logfile_handle.close()

    def sanitise(self, message: str) -> str:
        return "".join(ch for ch in message if unicodedata.category(ch)[0] != "C")

    def maybe_prefix(self, message: str, attributes: Dict[str, str]) -> str:
        if "machine" in attributes:
            return "{}: {}".format(attributes["machine"], message)
        return message

    def log_line(self, message: str, attributes: Dict[str, str]) -> None:
        self.xml.startElement("line", attributes)
        self.xml.characters(message)
        self.xml.endElement("line")

    def info(self, *args, **kwargs) -> None:  # type: ignore
        self.log(*args, **kwargs)

    def warning(self, *args, **kwargs) -> None:  # type: ignore
        self.log(*args, **kwargs)

    def error(self, *args, **kwargs) -> None:  # type: ignore
        self.log(*args, **kwargs)
        sys.exit(1)

    def log(self, message: str, attributes: Dict[str, str] = {}) -> None:
        self._eprint(self.maybe_prefix(message, attributes))
        self.drain_log_queue()
        self.log_line(message, attributes)

    def log_serial(self, message: str, machine: str) -> None:
        self.enqueue({"msg": message, "machine": machine, "type": "serial"})
        if self._print_serial_logs:
            self._eprint(
                Style.DIM + "{} # {}".format(machine, message) + Style.RESET_ALL
            )

    def enqueue(self, item: Dict[str, str]) -> None:
        self.queue.put(item)

    def drain_log_queue(self) -> None:
        try:
            while True:
                item = self.queue.get_nowait()
                msg = self.sanitise(item["msg"])
                del item["msg"]
                self.log_line(msg, item)
        except Empty:
            pass

    @contextmanager
    def nested(self, message: str, attributes: Dict[str, str] = {}) -> Iterator[None]:
        self._eprint(self.maybe_prefix(message, attributes))

        self.xml.startElement("nest", attrs={})
        self.xml.startElement("head", attributes)
        self.xml.characters(message)
        self.xml.endElement("head")

        tic = time.time()
        self.drain_log_queue()
        yield
        self.drain_log_queue()
        toc = time.time()
        self.log("(finished: {}, in {:.2f} seconds)".format(message, toc - tic))

        self.xml.endElement("nest")


rootlog = Logger()



import os
import socket
import click
import subprocess
import json
import time
from halo import Halo
from .actions import nix_store_location


class Platform(object):
    def __init__(self, ctx, name):
        self.name = name
        self.copy_from_store = False

    def retrieve_machines(self, ctx):
        pass

    # def get_start_values(self, ctx):
    #    pass


class Grid5000Platform(Platform):
    def __init__(self, ctx):
        super().__init__(ctx, "Grid5000")
        self.default_flavour = "g5k-nfs-store"
        self.copy_from_store = True
        self.first_start_values = ("ssh -l root", "sudo", None)
        self.subsequent_start_values = ("ssh -l root", "sudo", "/")
        self.oar_job_id = None
        self.oar_job = None
        self.group_users = "g5k-users"
        self.nix_store = nix_store_location(ctx)

    def retrieve_machines(self, ctx):
        oar_job = None
        oar_job_id_str = None
        # oar_job_id_prev_str = None
        halo = False

        # if "oar_job_id" in ctx.state:
        #    oar_job_id_prev_str = str(ctx.state["oar_job_id"])

        def oarstat():
            output = subprocess.check_output("oarstat -u -J", shell=True)
            if not output:
                raise click.ClickException(
                    click.style("Not oar job is present for this user", fg="red")
                )
            ctx.vlog(output)
            return json.loads(output)

        o = oarstat()
        if "OAR_JOB_ID" in os.environ:
            oar_job_id_str = os.environ["OAR_JOB_ID"]
        else:
            # we get the last one
            oar_job_ids = [int(jid) for jid in o.keys()]
            oar_job_ids.sort()
            if oar_job_ids:
                oar_job_id_str = str(oar_job_ids[-1])

        if not oar_job_id_str:
            raise click.ClickException(
                click.style("Unable to find a OAR_JOB_ID candidate", fg="red")
            )

        ctx.log(f"target OAR_JOB_ID={oar_job_id_str}")

        # TODO test when job exist but not Running or Launching (other states not supported)
        if oar_job_id_str in o.keys():
            oar_job = o[oar_job_id_str]
            if oar_job["state"] not in ["Running", "Launching", "Waiting"]:
                raise click.ClickException(
                    click.style(
                        f"fail to acquire OAR job {oar_job_id_str} in state: {oar_job['state']}",
                        fg="red",
                    )
                )

        if oar_job["state"] != "Running":
            spinner = Halo(text=f"Waiting OAR job: {oar_job['Job_Id']}", spinner="dots")
            halo = True
            spinner.start()

        while not oar_job or oar_job["state"] != "Running":
            time.sleep(0.25)
            o = oarstat()
            if oar_job_id_str in o.keys():
                oar_job = o[oar_job_id_str]

        if halo:
            spinner.succeed("OAR job: {oar_job['Job_Id']} id running")

        self.oar_job_id = int(oar_job_id_str)
        self.oar_job = oar_job

        # for oarsh but alter env
        os.environ["OAR_JOB_ID"] = oar_job_id_str

        return oar_job["assigned_network_address"]


# class LocalPlatform(Platform):
#     def __init__(self):
#         Platform.__init__(self, 'local')


def platform_detection(ctx):
    click.echo("\nPlatform detection:")
    split_hostname = socket.gethostbyaddr(socket.gethostname())[0].split(".")
    if len(split_hostname) >= 3 and split_hostname[2] == "grid5000":
        plt = click.style("   Grid'5000", fg="green")
        click.echo("   " + plt + " detected")
        ctx.platform = Grid5000Platform(ctx)
    else:
        click.echo("      no particular platform detected, local mode will be used")
    return



import time
from typing import List
from .driver.machine import Machine


class Flavour(object):
    name: str
    external_connect: bool = False
    machines: List[Machine] = []

    def __init__(self, ctx):
        self.ctx = ctx

    def check(self, state="running"):
        self.ctx.wlog(f"Check not implement for flavour: {self.name}")
        return -1

    def wait_on_check(self, state="running", mode="all", period=0.5, round=5):
        for _ in range(round):
            if mode == "all" and self.check(state) == len(self.machines):
                return True
            elif mode == "any" and self.check(state) > 0:
                return
            time.sleep(period)
        return False

    def generate_deployment_info(self, ssh_pub_key_file):
        pass

    def ext_connect(self, user, node, execute, ssh_key_file):
        pass


base_flavours = [
    {"name": "docker", "description": "Docker-Compose based", "image": {}},
    {
        "name": "g5k-image",
        "description": "Flavour for Grid'5000 platform",
        "image": {"distribution": "all-in-one", "type": "tarball"},
    },
    {
        "name": "g5k-nfs-store",
        "description": "Flavour for Grid'5000 platform",
        "image": {"distribution": "all-in-one", "type": "remote-store"},
    },
    {
        "name": "g5k-ramdisk",
        "description": "Flavour for Grid'5000 platform",
        "image": {"distribution": "all-in-one", "type": "ramdisk"},
    },
    {"name": "nspawn", "description": "Systemd-nspawn", "image": {}},
    {"name": "vm", "description": "vm", "image": {"distribution": "all-in-one"}},
    {
        "name": "vm-ramdisk",
        "description": "Plain vm ramdisk (all-in-memory), need lot of ram !",
        "image": {"distribution": "all-in-one", "type": "ramdisk"},
    },
]



import os
import os.path as op
import shutil
import sys
import subprocess
import click
import json

from ..actions import (
    get_nix_command,
    realpath_from_store,
    artifact_copy_all_kernel_initrd,
)
from ..context import pass_context, on_started, on_finished
from ..platform import platform_detection
from ..setup import apply_setup
from ..flavour import base_flavours

# FLAVOURS_PATH = op.abspath(op.join(op.dirname(__file__), "../", "flavours"))
# FLAVOURS = os.listdir(FLAVOURS_PATH)


@click.command("build")
@click.argument(
    "composition_file", required=False, type=click.Path(exists=True, resolve_path=True)
)
@click.option(
    "--nix-flags",
    type=click.STRING,
    help='add nix flags (aka options) to nix build command, --nix-flags "--impure"',
)
@click.option(
    "--namespace",
    "-N",
    type=click.STRING,
    help=(
        "Attribute namespace for compositions (e.g. `.#legacyPackages.x86_64-linux.hello` for compositions under `.#legacyPackages.x86_64-linux.hello.\"composition::<flavor>\"`)."
        "Defaults to `.#packages.x86_64-linux`"
    ),
)
@click.option("--out-link", "-o", help="path of the symlink to the build result")
@click.option(
    "-f",
    "--flavour",
    type=click.STRING,
    help="Use particular flavour (name or path)",
)
@click.option(
    "-F",
    "--list-flavours",
    is_flag=True,
    help="List available flavour",
)
@click.option(
    "-Fb",
    "--list-base-flavours",
    is_flag=True,
    help="List available base flavour",
)
# TOREMOVE
# @click.option(
#    "--copy-from-store",
#    "-c",
#    is_flag=True,
#    help="Copy artifacts (initrd, kernels, ...) from Nix store to artifact directory",
# )
@click.option("--show-trace", is_flag=True, help="Show Nix trace")
@click.option(
    "--dry-run", is_flag=True, help="Show what this command would do without doing it"
)
@click.option(
    "--dry-build",
    is_flag=True,
    help="Eval build expression and show store entry without building derivation",
)
@click.option(
    "-C",
    "--composition-flavour",
    type=click.STRING,
    help="Use to specify which composition and flavour combination to build when multiple compositions are describe at once (see -L options to list them).",
)
@click.option(
    "-L",
    "--list-compositions-flavours",
    is_flag=True,
    help="List available combinations of compositions and flavours",
)
@click.option(
    "-s",
    "--setup",
    type=click.STRING,
    help="Select setup variant",
)
@click.option(
    "-p",
    "--setup-param",
    type=click.STRING,
    multiple=True,
    help="Override setup parameter",
)
@click.option(
    "-u",
    "--update-flake",
    is_flag=True,
    help="Update flake.lock equivalent to: nix flake update",
)
@click.option(
    "--monitor",
    is_flag=True,
    help="Build with nix-output-monitor",
)
@click.option(
    "--mounted-store-url",
    "--mu",
    type=click.STRING,
    help="Use of nix experimental SSH store with filesystem mounted, format: [username@]hostname",
)
@pass_context
@on_finished(lambda ctx: ctx.show_elapsed_time())
@on_started(lambda ctx: ctx.assert_valid_env())
def cli(
    ctx,
    composition_file,
    nix_flags,
    namespace,
    out_link,
    flavour,
    list_flavours,
    list_base_flavours,
    show_trace,
    dry_run,
    dry_build,
    composition_flavour,
    list_compositions_flavours,
    update_flake,
    setup,
    setup_param,
    monitor,
    mounted_store_url,
):
    """
    Builds the composition.

    It generates a `build` folder which stores symlinks to the closure associated to a composition. The file name of the symlink follows this structure  `[composition-name]::[flavour]`

    ## Examples

    - `nxc build -f vm`

        Build the `vm` flavour of your composition.

    - `nxc build -C oar::g5k-nfs-store`

        Build the `oar` composition with the `g5k-nfs-store` flavour.
    """

    def determine_flavour(ctx):
        if "default_flavour" in ctx.nxc and ctx.nxc["default_flavour"]:
            flavour = ctx.nxc["default_flavour"]
        else:
            platform_detection(ctx)
            if ctx.platform:
                flavour = ctx.platform.default_flavour
            else:
                flavour = "default"
        ctx.vlog(f"Selected flavour: {flavour}")
        return flavour

    def namespace_attr_path(value):
        if "#" in value:
            attr = value.split("#", 1)[1]
        else:
            attr = value
        return attr.split(".") if attr else []

    def namespace_flake_ref(value):
        if "#" not in value:
            return None
        ref = value.split("#", 1)[0]
        if not ref or ref == ".":
            return None
        return ref

    if not namespace:
        namespace = ".#packages.x86_64-linux"
    ctx.namespace = namespace

    if setup and not op.exists(op.join(ctx.envdir, "setup.toml")):
        ctx.elog("setup option is given but setup.toml is not found")
        sys.exit(1)

    if setup or op.exists(op.join(ctx.envdir, "setup.toml")):
        nix_flags, composition_file, composition_flavour, flavour, _ = apply_setup(
            ctx,
            setup,
            nix_flags,
            composition_file,
            composition_flavour,
            flavour,
            setup_param,
            None,
        )

    build_cmd = []

    # Do we are in flake context
    if not op.exists(op.join(ctx.envdir, "flake.nix")):
        ctx.elog("Not Found flake.nix file")
        sys.exit(1)

    if monitor:
        nix_cmd_base = ["nom"]
    else:
        nix_cmd_base = get_nix_command(ctx)

    if update_flake:
        cmd = nix_cmd_base + ["flake", "update"]
        if ctx.show_spinner:
            ctx.spinner.start("Updating flake.lock")
            ret = subprocess.call(
                cmd,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                cwd=ctx.envdir,
            )
            if ret:
                ctx.spinner.stop()
                ctx.elog("Updating flake.lock is failed")
                sys.exit(1)
            else:
                ctx.spinner.succeed("Updating flake.lock is done")

    description_flavours = get_flavours(nix_cmd_base, ctx)

    flavours = list(description_flavours.keys())

    if list_flavours:
        ctx.log("Flavours List:")
        for k in flavours:
            click.echo(f"{k: <18}: {description_flavours[k]['description']}")
        sys.exit(0)

    if list_base_flavours:
        flavours = get_base_flavours()
        for flavour in flavours:
            click.echo(f"{flavour['name']: <18}: {flavour['description']}")
        sys.exit(0)

    if not composition_file:
        composition_file = ctx.nxc["composition"]

    if list_compositions_flavours:
        cmd = nix_cmd_base + ["flake", "show", "--json"]
        flake_ref = namespace_flake_ref(namespace)
        if flake_ref:
            cmd += [flake_ref]
        attr_path = namespace_attr_path(namespace)
        if attr_path and attr_path[0] == "legacyPackages":
            cmd += ["--legacy"]
        raw_compositions_flavours = json.loads(
            subprocess.check_output(cmd, cwd=ctx.envdir).decode()
        )
        if not attr_path:
            raise click.ClickException(f"Namespace is empty: {namespace}")
        current = raw_compositions_flavours
        for part in attr_path:
            if not isinstance(current, dict) or part not in current:
                raise click.ClickException(
                    f"Namespace not found in flake outputs: {namespace}"
                )
            current = current[part]
        if not isinstance(current, dict):
            raise click.ClickException(
                f"Namespace does not point to an attribute set: {namespace}"
            )
        for compo_flavour in filter(
            lambda x: x not in ["flavoursJson", "showFlavours"],
            current.keys(),
        ):
            print(compo_flavour)
        if namespace == ".#packages.x86_64-linux":
            print(
                click.style("Default", fg="green")
                + ": "
                + raw_compositions_flavours["defaultPackage"]["x86_64-linux"]["name"]
            )
        sys.exit(0)

    if show_trace:
        build_cmd += ["--show-trace"]

    if mounted_store_url:
        build_cmd = [
            "--extra-experimental-features",
            "mounted-ssh-store",
            "--store",
            f"mounted-ssh-ng://{mounted_store_url}",
        ]

    if flavour and composition_flavour:
        if len(composition_flavour.split("::")) == 1:
            composition_flavour = composition_flavour + "::" + flavour

    if not out_link:
        build_path = op.join(ctx.envdir, "build")
        if not op.exists(build_path):
            create = click.style("   create", fg="green")
            ctx.log("   " + create + "  " + build_path)
            os.mkdir(build_path)

        if composition_flavour:
            if flavour and flavour != composition_flavour.split("::")[-1]:
                raise ValueError(
                    "the value of flavour  does not match  the ones of composition_favour"
                )

            ctx.composition_flavour_prefix = composition_flavour
            ctx.flavour_name = composition_flavour[-1]

        else:
            composition_name = (os.path.basename(composition_file)).split(".")[0]
            ctx.composition_name = composition_name
            if not flavour:
                flavour = determine_flavour(ctx)
            ctx.flavour_name = flavour
            ctx.composition_flavour_prefix = f"{composition_name}::{flavour}"

        out_link = op.join(build_path, ctx.composition_flavour_prefix)

    if not flavour:
        flavour = determine_flavour(ctx)

    if not composition_flavour:
        composition_flavour = f"composition::{flavour}"

    if dry_build:
        build_cmd = nix_cmd_base + ["eval"] + build_cmd + ["--raw"]
    else:
        build_cmd = nix_cmd_base + ["build"] + build_cmd
        if out_link and not mounted_store_url:
            build_cmd += ["-o", out_link]
        if mounted_store_url:
            build_cmd += ["--no-link", "--json"]

    # add additional nix flags if any
    if nix_flags:
        build_cmd += nix_flags.split()

    build_cmd += [f"{namespace}.{composition_flavour}"]

    if not dry_run:
        ctx.glog("Starting Build")
        ctx.vlog(build_cmd)
        if mounted_store_url:
            proc = subprocess.run(build_cmd, cwd=ctx.envdir, stdout=subprocess.PIPE)
            returncode = proc.returncode
            if not returncode:
                nix_build_output = json.loads(proc.stdout)
                # create link here and not by nix, to avoid issue directory is not accessible with remote build .
                ctx.vlog(f"nix build output: {nix_build_output}")
                if os.path.exists(out_link):
                    os.remove(out_link)
                shutil.copyfile(nix_build_output[0]["outputs"]["out"], out_link)

                if flavour == "g5k-nfs-store":
                    ctx.compose_info_file = out_link
                    artifact_copy_all_kernel_initrd(ctx)
        else:
            returncode = subprocess.call(build_cmd, cwd=ctx.envdir)
        if returncode:
            ctx.elog(f"Build return code: {returncode}")
            sys.exit(returncode)

        # Loading the docker image
        # todo: to move in docker flavour class
        if flavour == "docker" and not dry_build:
            out_link = realpath_from_store(ctx, out_link)
            with open(out_link, "r") as compose_info_json:
                content = json.load(compose_info_json)
                docker_image = realpath_from_store(ctx, content["image"])
                docker_load_command = f"docker load < {docker_image}"
                returncode = subprocess.call(docker_load_command, shell=True)
                if returncode:
                    ctx.elog(f"Build return code: {returncode}")
                    sys.exit(returncode)
            ctx.glog("Docker Image loaded")

        ctx.glog("\nBuild completed")
    else:
        ctx.log("Dry-run:")
        ctx.log(f"   working directory:          {ctx.envdir}")
        ctx.log(f"   composition flavour prefix: {ctx.composition_flavour_prefix}")
        ctx.log(f"   build command:              {' '.join(build_cmd)}")


def get_flavours(nix_cmd_base, ctx):
    """
    Returns the json representation of the available flavours
    """
    FLAVOURS_JSON = op.abspath(
        op.join(op.dirname(__file__), "../../nix", "flavours.json")
    )

    output_json = FLAVOURS_JSON

    # TODO add option to build flavours list from nix
    #
    # flake_location = "."
    # output_json = "/tmp/.flavours.json"
    # ctx.log("Build list of flavours")
    # retcode = subprocess.call(
    #     nix_cmd_base + ["build", f"{flake_location}#flavoursJson", "-o", output_json],
    #     stdout=subprocess.DEVNULL,
    #     stderr=subprocess.DEVNULL,
    # )
    # if retcode:
    #     output_json = FLAVOURS_JSON
    # else:
    #     if not op.exists(output_json):
    #         output_json = realpath_from_store(ctx, output_json)

    return json.load(open(output_json, "r"))


def get_base_flavours():
    return base_flavours



import os
import os.path as op
import subprocess
import click
import ipaddress
import socket

from ..flavour import Flavour
from ..actions import (
    read_compose_info,
    read_deployment_info,
    generate_deployment_info,
)

from ..driver.machine import Machine

# from ..default_role import DefaultRole

from typing import Tuple, Optional


def nft_nixos_fw_rules(ctx, remove=False, add=False):
    subprocess.call("sudo true", shell=True)
    check_process = ""
    try:
        check_process = subprocess.check_output(
            [
                "sudo",
                "nft",
                "--handle",
                "--numeric",
                "list",
                "chain",
                "ip",
                "filter",
                "nixos-fw",
            ],
            stderr=subprocess.DEVNULL,
        )

    except subprocess.CalledProcessError as e:
        ctx.wlog(f"nftable: ip chain filter not present, return code {e.returncode}")
        # print(f"{e.output} return code {e.returncode}")
        return -1  # non present

    nxc_br0_rule = False
    nxc_br0_rule_handle = -1
    for line in check_process.decode().rstrip("\n").splitlines():
        if 'iifname "nxc-br0"' in line:
            nxc_br0_rule = True
            s = line.split()
            if s[-2] == "handle":
                nxc_br0_rule_handle = s[-1]

    if remove and nxc_br0_rule:
        subprocess.Popen(
            [
                "sudo",
                "nft",
                "delete",
                "rule",
                "ip",
                "filter",
                "nixos-fw",
                "handle",
                nxc_br0_rule_handle,
            ]
        )
        return 1
    if add and not nxc_br0_rule:
        subprocess.Popen(
            [
                "sudo",
                "nft",
                "insert",
                "rule",
                "ip",
                "filter",
                "nixos-fw",
                'iifname "nxc-br0" counter packets 0 bytes 0 accept',
            ]
        )
        return 1
    return 0


def set_prefix_store_volumes(dc_json, prefix_store):
    for service in dc_json["services"].keys():
        volumes = dc_json["services"][service]["volumes"]
        volumes_out = []
        for vol in volumes:
            if "/nix/store" == vol[:10]:
                volumes_out.append(f"{prefix_store}{vol[4:]}")
            else:
                volumes_out.append(vol)
        dc_json["services"][service]["volumes"] = volumes_out


class NspawnFlavour(Flavour):
    nspawn_compose_file = None

    def __init__(self, ctx):
        super().__init__(ctx)

        self.name = "nspawn"
        self.description = "Systemd-nspawn container"

    @staticmethod
    def host_info(role, hostname, info):
        return {"role": role, "host": hostname, "toplevel": info["toplevel"]}

    def generate_deployment_info(self, ssh_pub_key_file):
        ctx = self.ctx
        deploy_dir = op.join(ctx.envdir, "deploy")
        if not op.exists(deploy_dir):
            create = click.style("   create", fg="green")
            ctx.log("   " + create + "  " + deploy_dir)
            os.mkdir(deploy_dir)

        if not ctx.ip_range:
            ctx.ip_range = "10.0.3.2,10.0.3.254"

        ip_l, ip_h = ctx.ip_range.split(",")
        subnets = ipaddress.summarize_address_range(
            ipaddress.IPv4Address(ip_l), ipaddress.IPv4Address(ip_h)
        )
        ctx.ip_addresses = [str(ip) for s in subnets for ip in s]

        # TODO ugly need to harmonisation build/composition_info through all flavours
        if not ctx.compose_info:
            read_compose_info(ctx)
        ctx.compose_info["roles"] = ctx.compose_info[ctx.composition_name]

        generate_deployment_info(ctx, ssh_pub_key_file)

    def driver_initialize(self, tmp_dir):
        print("TODO driver_initialize")
        exit
        assert self.ctx.deployment_info
        if not self.nspawn_compose_file:
            self.nspawn_compose_file = self.ctx.deployment_info["nspawn-compose-file"]

        nodes_names = self.ctx.deployment_info["nodes"]
        for name in nodes_names:
            self.machines.append(
                Machine(
                    self.ctx,
                    tmp_dir=tmp_dir,
                    start_command="",
                    name=name,
                )
            )

    def check(self, state="running"):
        print("TODO check")
        exit
        # check_process = subprocess.check_output(
        #     [
        #         "nspawn-compose",
        #         "-f",
        #         self.nspawn_compose_file,
        #         "ps",
        #         "--services",
        #         "--filter",
        #         f"status={state}",
        #     ],
        # )
        # return len(check_process.decode().rstrip("\n").splitlines())

    def connect(self, machine):
        if machine.connected:
            return
        self.start_all()

    def launch(self):
        ctx = self.ctx
        if not ctx.deployment_info:
            read_deployment_info(ctx)

        nest_host = ""

        if "nested" in ctx.deployment_info and ctx.deployment_info["nested"]:
            nest_host = f"-{socket.gethostname()}"

        # prepare nxc-dnsmasq.conf
        artifact_dir = op.join(
            ctx.envdir, f"artifact/{ctx.composition_name}/{ctx.flavour.name}"
        )
        os.makedirs(artifact_dir, mode=0o700, exist_ok=True)
        nxc_dnsmasq_conf_file = op.join(artifact_dir, f"nxc-dnsmasq{nest_host}.conf")

        with open(nxc_dnsmasq_conf_file, "w") as outfile:
            for ip, host_info in ctx.deployment_info["deployment"].items():
                outfile.write(f"dhcp-host={host_info['host']},{ip}\n")

        _ROOT = os.path.abspath(os.path.dirname(__file__))
        nxc_net_script = _ROOT + "/nspawn/nxc-net.sh"
        machine_dir_script = _ROOT + "/nspawn/machine-dir.sh"

        env = os.environ

        ctx.log("Prepare and launch nspawn container")
        ctx.log("Test sudo (root privilege rights required)")
        subprocess.call("sudo true", shell=True)

        env["NXC_DHCP_CONFILE"] = nxc_dnsmasq_conf_file
        preserve_env = "NXC_DHCP_CONFILE"

        if "nested" in ctx.deployment_info and ctx.deployment_info["nested"]:
            nested_network = ctx.deployment_info["network"]
            net_addr, net_masq = nested_network.split("/")
            if net_masq != "24":
                ctx.elog("Netmask different from /24 is not supported")
            a, b, c, _ = net_addr.split(".")
            net_prefix = f"{a}.{b}.{c}"
            env["NXC_ADDR"] = f"{net_prefix}.1"
            env["NXC_NETWORK"] = nested_network
            env["NXC_DHCP_RANGE"] = f"{net_prefix}.2,{net_prefix}.254"
            preserve_env += ",NXC_ADDR,NXC_NETWORK,NXC_DHCP_RANGE"

        ctx.log("Launch nxc-net script")

        subprocess.Popen(
            ["sudo", f"--preserve-env={preserve_env}", nxc_net_script, "start"],
            env=env,
        )

        ctx.log("Check and adapt nftable nixos-fw chain if any")

        nft_nixos_fw_rules(ctx, add=True)

        ctx.log("Prepare machines dirs")
        p_lst = []
        for _, host_info in ctx.deployment_info["deployment"].items():
            p = subprocess.Popen(
                [
                    "sudo",
                    machine_dir_script,
                    "prepare",
                    host_info["host"],
                    host_info["toplevel"],
                ],
                stdout=subprocess.DEVNULL,
            )
            p_lst.append(p)

        for p in p_lst:
            p.wait()

        ctx.log("\nStart containers")
        p_lst = []

        env["SYSTEMD_NSPAWN_UNIFIED_HIERARCHY"] = "1"

        for _, host_info in ctx.deployment_info["deployment"].items():
            subprocess.Popen(
                [
                    "sudo",
                    "--preserve-env=SYSTEMD_NSPAWN_UNIFIED_HIERARCHY",
                    "systemd-nspawn",
                    "-bD",
                    f"/var/lib/machines/{host_info['host']}",
                    "--network-bridge=nxc-br0",
                ],
                env=env,
                # stdout=subprocess.DEVNULL,
                # stderr=subprocess.DEVNULL,
            )
            p_lst.append(p)
        for p in p_lst:
            p.wait()

    def start_all(self):
        print("TODO start_all")
        exit
        pass
        # if not self.external_connect:
        #     with rootlog.nested("starting nspawn-compose"):
        #         subprocess.Popen(
        #             ["nspawn-compose", "-f", self.nspawn_compose_file, "up", "-d"]
        #         )

        #     self.wait_on_check()

        # for machine in self.machines:
        #     if not machine.connected:
        #         self.start(machine)
        #         machine.connected = True

    def start(self, machine):  # TODO MOVE to Connect ???
        print("TODO start")
        exit
        pass
        assert machine.name
        assert self.nspawn_compose_file

        # machine.start_process_shell(
        #     [
        #         "nspawn-compose",
        #         "-f",
        #         self.nspawn_compose_file,
        #         "exec",
        #         "-u",
        #         "root",
        #         "-T",
        #         machine.name,
        #         "bash",
        #         "-l",
        #     ]
        # )

    def execute(
        self,
        machine,
        command: str,
        check_return: bool = True,
        timeout: Optional[int] = 900,
    ) -> Tuple[int, str]:
        return machine.execute_process_shell(command, check_return, timeout)

    def restart(self, machine):
        machine.restart_process_shell()

    def cleanup(self):
        ctx = self.ctx
        if not ctx.deployment_info:
            read_deployment_info(ctx)

        _ROOT = os.path.abspath(os.path.dirname(__file__))
        nxc_net_script = _ROOT + "/nspawn/nxc-net.sh"
        machine_dir_script = _ROOT + "/nspawn/machine-dir.sh"

        ctx.log("Stop and cleanup nspawn container(s)")
        ctx.log("Test sudo (root privilege rights required)")
        subprocess.call("sudo true", shell=True)

        for _, host_info in ctx.deployment_info["deployment"].items():
            subprocess.Popen(
                [
                    "sudo",
                    "machinectl",
                    "stop",
                    host_info["host"],
                ]
            )

        for _, host_info in ctx.deployment_info["deployment"].items():
            subprocess.Popen(
                [
                    "sudo",
                    machine_dir_script,
                    "remove",
                    host_info["host"],
                ]
            )
        ctx.log("Stop nxc-net")

        subprocess.Popen(["sudo", nxc_net_script, "stop"])
        ctx.log("Remove nftable chain nixos-fw if any")
        nft_nixos_fw_rules(ctx, remove=True)

    def shell_interact(self, machine) -> None:
        self.connect(machine)
        self.ext_connect("root", machine.name)

    def ext_connect(self, user, node, execute=True, ssh_key_file=None):
        # subprocess.call("sudo true", shell=True)
        cmd = f"machinectl shell {user}@{node}"
        if execute:
            return_code = subprocess.run(cmd, shell=True).returncode

            if return_code:
                self.ctx.wlog(f"Machinectl exit code is not null: {return_code}")
            return return_code
        else:
            return cmd



"""
   nixos_compose.flavours
   Adapted from pygments project (pygments.styles)
   :copyright: Copyright 2006-2021 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

#: Maps flavour names to 'submodule::classname'.
FLAVOUR_MAP = {
    "docker": "docker::DockerFlavour",
    "nspawn": "nspawn::NspawnFlavour",
    "vm": "vm::VmFlavour",
    "vm-ramdisk": "vm::VmRamdiskFlavour",
    "g5k-nfs-store": "grid5000::G5kNfsStoreFlavour",
    "g5k-ramdisk": "grid5000::G5kRamdiskFlavour",
    "g5k-image": "grid5000::G5kImageFlavour",
}


class ClassNotFound(ValueError):
    """Raised if one of the lookup functions didn't find a matching class."""


def get_flavour_by_name(name):
    if name in FLAVOUR_MAP:
        mod, cls = FLAVOUR_MAP[name].split("::")

    try:
        mod = __import__("nixos_compose.flavours." + mod, None, None, [cls])
    except ImportError:
        raise ClassNotFound(f"Could not find flavour module {mod}")
    try:
        return getattr(mod, cls)
    except AttributeError:
        raise ClassNotFound(f"Could not find flavour class {cls} in flavour module.")


def use_flavour_method_if_any(f):
    def wrapper(*args, **kwargs):
        attr_name = f.__name__
        flavour = args[0].ctx.flavour
        if hasattr(flavour, attr_name):
            g = getattr(flavour, attr_name)
            # isinstance isn't use to avoid circular dependencies
            if args[0].__class__.__name__ == "Driver":
                args = args[1:]
            return g(*args, **kwargs)
        else:
            return f(*args, **kwargs)

    return wrapper



#!/usr/bin/env python
from execo import Remote
from execo_engine import Engine, logger
from execo_g5k import (
    OarSubmission,
    oardel,
    oarsub,
    wait_oar_job_start,
)
from nixos_compose.nxc_execo import get_oar_job_nodes_nxc


class MyEngine(Engine):
    def __init__(self):
        super(MyEngine, self).__init__()
        self.oar_job_id = None
        parser = self.args_parser
        parser.add_argument("--nxc_build_file", help="Path to the NXC build file")
        parser.add_argument("--walltime", help="Grid5000 booking walltime (in hours)")
        parser.add_argument(
            "--flavour", help="Nixos compose flavour", default="g5k-image"
        )
        parser.add_argument(
            "--keep-job",
            help="Do not delete the OAR job after execution, be it after an error or success",
            action="store_true",
        )
        parser.add_argument(
            "--job-id",
            help="When provided, the given OAR job ID is used and no further booking is done",
            type=int,
        )

    def init(self):
        pass

    def run(self):
        # Initialise some experiment parameters
        site = "grenoble"
        cluster = "dahu"
        nxc_flavour = self.args.flavour
        nb_nodes = 2
        roles_distribution = {
            "foo": ["foo", "bar"],
        }
        walltime_hours = float(self.args.walltime) if self.args.walltime else 1
        # Local copies of experiment parameters

        try:
            # Book nodes on Grid 5000 unless the ID of an existing job has been provided
            if self.args.job_id is None:
                # Book nodes on Grid'5000
                logger.info(f"Reserving {nb_nodes} node.s on {site}-{cluster}...")
                oar_job = reserve_nodes(
                    nb_nodes, site, cluster, "deploy", walltime=walltime_hours * 60 * 60
                )
                self.oar_job_id, site = oar_job[0]

                logger.info(f"Waiting for job ID {self.oar_job_id} on {site} site...")
                wait_oar_job_start(self.oar_job_id, site)
            else:
                self.oar_job_id = self.args.job_id

            # Get the machines info, deploy
            logger.info("Deploying ...")
            nodes, roles = get_oar_job_nodes_nxc(
                self.oar_job_id,
                site,
                flavour_name=nxc_flavour,
                compose_info_file=self.args.nxc_build_file,
                # composition_name = "", # for multiple compositions case
                roles_quantities=roles_distribution,
            )
            logger.info(f"... done. Nodes used for this experiment: {nodes}")

            #
            logger.debug("Execute on hello")

            my_command = 'echo "Hello from $(whoami) at $(hostname) ($(ip -4 addr | grep "/20" | awk \'{print $2;}\'))" > /tmp/hello'
            hello_remote = Remote(
                my_command, nodes["foo"], connection_params={"user": "root"}
            )
            hello_remote.run()
            # throw_on_problem( hello_remote)

            my_command2 = "cat /tmp/hello"
            cat_remote = Remote(
                my_command2, nodes["foo"], connection_params={"user": "root"}
            )
            cat_remote.run()
            for process in cat_remote.processes:
                print(process.stdout)
        except FailedProcessError as e:
            logger.error(f"Failed at process {e}")
        except KeyboardInterrupt:
            logger.info("Stopping (received keyboard interrupt)")
        finally:
            if self.oar_job_id is not None and not self.args.keep_job:
                logger.info(f"Giving back the resources (OAR job ID {self.oar_job_id})")
                oardel([(self.oar_job_id, "site")])


class FailedProcessError(Exception):
    """Exception raised when an Execo process failed (meaning its attribute finished_ok is False)"""

    def __init__(self, process):
        super(Exception, self).__init__()
        self.process = process

    def __str__(self):
        return str(self.process)


def throw_on_problem(process):
    """Raise if the given process did not finish properly (meaning attribute finished_ok is False)"""
    if not process.finished_ok:
        raise FailedProcessError(process)


def reserve_nodes(nb_nodes, site, cluster, job_type, walltime=3600):
    """
    :param walltime: the duration of the job, in seconds (or a datetime, or a string as expected by the oarsub program)
    """
    jobs = oarsub(
        [
            (
                OarSubmission(
                    resources="{{cluster='{}'}}/nodes={}".format(cluster, nb_nodes),
                    walltime=walltime,
                    job_type=[job_type],
                ),
                # additional_options = '-t exotic'),
                site,
            )
        ]
    )
    return jobs


if __name__ == "__main__":
    ENGINE = MyEngine()
    try:
        ENGINE.start()
    except Exception as ex:
        print(f"Failing with error {ex}")
        oardel([(ENGINE.oar_job_id, None)])
        print("Giving back the resources")
